from . import parse_subroutine, parse_function
from .code import Code

# PROCEDURE cannot have "INTENT".
# 
# All passed procedure arguments, will be defined as 
#   'ABSTRACT INTERFACE's and declared at the front of the 
#   definitions for a routine. Since they are passed, they must
#   be procedures with the BIND(C) attribute enabled.


# For parsing purposes, an Interface looks like a Module.
class Interface(Code):
    type = "INTERFACE"
    named_end = False
    # prefixes = ["ABSTRACT"] # <- need another type for abstrac ones.
    can_contain = [(parse_subroutine, "subroutines"),
                   (parse_function, "functions")
    ]

    # Produce a string representation of this Subroutine object.
    def __str__(self):
        # Print out header line.
        out =  "INTERFACE\n"
        for f in self.subroutines:
            for line in str(f).split("\n"):
                out += "  "+line+"\n"
        if (len(self.subroutines) > 0): out += "\n"
        for f in self.functions:
            for line in str(f).split("\n"):
                out += "  "+line+"\n"
        out += "END INTERFACE"
        # Return the whole interface.
        return out

    # Given a list of lines (of a source Fortran file), parse out this
    # Function (assuming the first line is the first line *inside* of
    # this Function).
    def parse(self, list_of_lines):
        # Pop out the first line, it's not needed.
        list_of_lines.pop(0)
        # Use standard parsing code (looks for subroutines and functions).
        super().parse(list_of_lines)



# "util" should perform compilation of fortran codes in setup, saving requires admin privaleges.
# Passed functions store python functions associated with random integer
#   identifier, that is used to look up correct function in cython.

# An example of how to create containers of structures in Numpy
#   https://stackoverflow.com/questions/50822183/efficient-conversion-from-c-array-of-struct-to-numpy-array-with-ctypes
#   https://stackoverflow.com/questions/40063080/casting-an-array-of-c-structs-to-a-numpy-array

from ctypes import Structure, c_char_p, c_int, c_double, POINTER
import numpy as np

def c_func(): pass

class info(Structure):
    _fields_ = [ ("name", c_char_p),
                 ("arr_len", c_int),
                 ("real_data", POINTER(c_double)) ]

c_func.restype = info
ret_val = c_func()
# data = np.ctypeslib.as_array(ret_val.contents.real_data



from .argument import Argument

COMPLEX_STRUCT = """
# This defines a C structure that can be used to hold the complex C type.
class c_complex_{type}(ctypes.Structure):
    _fields_ = [("real", ctypes.c_{type}), ("imag", ctypes.c_{type})]
    # Define an "__init__" that can take a complex number as input.
    def __init__(self, real=0.0, imag=None):
        # If the provided first value was a complex number, expand it out.
        if (hasattr(real, "imag") and hasattr(real, "real") and (imag is None)):
            imag = real.imag
            real = real.real
        # Assign the internal fields.
        self.real = real
        self.imag = imag
    # Define a "value" so that it can be cleanly retreived as a single complex number.
    @property
    def value(self):
        return complex(self.real, self.imag)
"""

COMPLEX_FLOAT = COMPLEX_STRUCT.format(type="float")
COMPLEX_DOUBLE = COMPLEX_STRUCT.format(type="double")
COMPLEX_LONG_DOUBLE = COMPLEX_STRUCT.format(type="longdouble")

class Complex(Argument):
    type = "COMPLEX"
    default_singleton = "1"
    c_types = {"32":"c_complex_double", "16":"c_complex_double", "8":"c_complex_float"}
    c_types_arrays = {"32":"numpy.complex128", "16":"numpy.complex128", "8":"numpy.complex64"}
    py_types = {"32":COMPLEX_DOUBLE, "16":COMPLEX_DOUBLE, "8":COMPLEX_FLOAT}

    @property
    def py_type(self):
        if (self.size not in self.py_types):
            raise(NotImplementedError(f"\n\nUnrecognized size '{self.size}' for argument '{self.name}', no known corresponding Pythonx type."))
        return self.py_types[self.size]






''''''


import cython
import numpy

class NotFortranCompatible(Exception): pass

#      Interface for fortran function quadratic     
# =================================================

cdef void c_quadratic(float* x, float* y):
    # Recall the python function from global 
    #  (this was assigned by cython during python wrapper call)
    global quadratic_global
    # Define any variables needed to copy arrays
    
    # Create local variables that are python types
    
    # Call function with python variables in and out
    y[0] = quadratic_global(x[0])
    # Copy out from python variables back into c-typed (fortran) variables
    

#      Wrapper for fortran function test_standard     
# =================================================

cdef extern:
    void c_test_standard( void (*quadratic)(float*, float*), int* steps, float* solution, float* max_window, bint* max_window_present )

@cython.boundscheck(False)
@cython.wraparound(False)
def test_standard( quadratic, int steps, float solution, max_window=None ):
    '''! Minimize a quadratic by taking random'''
    # Prepare for fortran function call (initialize optionals)
    global quadratic_global
    quadratic_global = quadratic
    cdef bint max_window_present = True
    if (type(max_window) == type(None)):
        max_window_present = False
        max_window = 1
    cdef float local_max_window = max_window
    
    
    # Make fortran function call
    c_test_standard(&c_quadratic, &steps, &solution, &local_max_window, &max_window_present)
    # Return appropriate values based on "INTENT" from fortran code
    
    return solution






# Add spaces to syntax in lines (for easier splitting by white-space)
FORT_TEXT_REPLACEMENTS = {
    "DOUBLE PRECISION": "REAL ( KIND ( 0.0D0 ) )",
    "END DO": "ENDDO",
    "END IF": "ENDIF",
    ",":" , ",
    "(":" ( ",
    ")":" ) ",
    ":":" : ",
    "+":" + ",
    "-":" - ",
    "*":" * ",
    "/":" / ",
    "=":" = ",
}

# Some spaces that were created above were misplaced, so this fixes them.
FORT_TEXT_FIXES = {
    "= >":"=>",
}

# Keep all comment lines and all lines from a file that start with the
# following expressions (the file is converted to all upper case).
ACCEPTABLE_LINE_STARTS = {'ABSTRACT', 'CHARACTER', 'END', # 'EXTERNAL',
                          'INTEGER', 'LOGICAL', 'REAL', 'COMPLEX',
                          'IMPLICIT', 'INTERFACE', 'MODULE', 'FUNCTION',
                          'OPTIONAL', 'PRIVATE', 'PROCEDURE',
                          'PUBLIC', 'PURE', 'RECURSIVE',
                          'SUBROUTINE', 'TYPE', 'USE'}
LINE_STARTS_TO_REMOVE = {'PURE', 'RECURSIVE'}

# Immediately exclude a file from automatic compilation if it has a
# line starting with the following.
IMMEDIATELY_EXCLUDE = {"PROGRAM"}

please_report_to = f"\nIf this is syntactically correct Fortran, please report this\n"+\
                   f"as an issue with the relevant Fortran code at\n\n"+\
                   f"  https://github.com/tchlux/fmodpy\n"


from .util import after_dot, before_dot, class_name, \
    legal_module_name, simplify_fortran_file, pop_group, \
    wrap_long_lines

# --------------------------------------------------------------------

# Define a function for parsing an Interface. Returns a list of instances
def parse_interface(list_of_lines, comments, parent):
    from .interface import Interface
    return parse_code(Interface, list_of_lines, comments, parent)

# Define a function for parsing a Module. Returns a list of instances
def parse_module(list_of_lines, comments, parent):
    from .module import Module
    return parse_code(Module, list_of_lines, comments, parent)

# Define a function for parsing a Subroutine. Returns a list of instances
def parse_subroutine(list_of_lines, comments, parent):
    from .subroutine import Subroutine
    return parse_code(Subroutine, list_of_lines, comments, parent)

# Define a function for parsing a Function. Returns a list of instances
def parse_function(list_of_lines, comments, parent):
    from .function import Function
    return parse_code(Function, list_of_lines, comments, parent)

# Define a function for parsing a Type. Returns a list of instances
def parse_type(list_of_lines, comments, parent):
    from .type import TypeDeclaration
    return parse_code(TypeDeclaration, list_of_lines, comments, parent)

# Parse a PUBLIC line, identifying routines that are public.
def parse_public(list_of_lines, comments, parent, keyword="PUBLIC"):
    # Skip if there are no lines to process.
    if (len(list_of_lines) == 0): return []
    line = list_of_lines[0].strip().split()
    # Remove any comments on the line that may exist.
    if ("!" in line): line = line[:line.index("!")]
    # Skip empty lines.
    if (len(line) == 0):
        list_of_lines.pop(0)
        return []
    # Check to see if the first element is 'PUBLIC'.
    if (line[0] == keyword):
        # If there is only the word 'PUBLIC', then it is a status.
        if (len(line) == 1):
            parent.status = keyword
            # Remove this line from the list (because it has been parsed).
            list_of_lines.pop(0)
            return []
        # Strip out the double colon if it exists (it's optional).
        if (''.join(line[1:3]) == "::"): line = line[3:]
        # Remove all commas from the line.
        commas = [i for i in range(len(line)) if (line[i] == ",")]
        for i in reversed(commas): line.pop(i)
        # Remove this line from the list (because it has been parsed).
        list_of_lines.pop(0)
        # Only variable and function names remain, they are the instances.
        return line
    # Otherwise, no 'PUBLIC' line found, return empty list.
    return []

# Parse a PRIVATE line, identifying routines that are private.
def parse_private(list_of_lines, comments, parent):
    return parse_public(list_of_lines, comments, parent, keyword="PRIVATE")

# Parse a USE line, return a string in a list.
def parse_use(list_of_lines, comments, parent):
    # Skip if there are no lines to process.
    if (len(list_of_lines) == 0): return []
    line = list_of_lines[0].strip().split()
    # Skip empty lines.
    if (len(line) == 0): return []
    # Check to see if the first element is "USE".
    # If it is, then return a list with one string (the line).
    if (line[0] == "USE"): return [list_of_lines.pop(0).replace(": :","::")]
    # Otherwise, no USE line found, return empty list.
    else: return []

# Parse a line with IMPLICIT NONE, return string in a list.
def parse_implicit(list_of_lines, comments, parent):
    # Skip if there are no lines to process.
    if (len(list_of_lines) == 0): return []
    line = list_of_lines[0].strip().split()
    # Skip empty lines.
    if (len(line) == 0): return []
    # Check to see if the first element is "IMPLICIT".
    # If it is, then return a list with one string (the line).
    if (line[0] == "IMPLICIT"):
        if (len(line) > 1):
            if (line[1] == "NONE"): return [list_of_lines.pop(0)]
        else:
            from fmodpy.exceptions import ParseError
            raise(ParseError("Found phrase 'IMPLICIT' but it was not followed by 'NONE'."))
    # Otherwise, no IMPLICIT line found, return empty list.
    else: return []

# Only initialize a "code" object if the line is not empty and the
# type is matched (after removing any prefixes).
def parse_code(code, list_of_lines, comments, parent):
    # Skip if there are no lines to process.
    if (len(list_of_lines) == 0): return []
    line = list_of_lines[0].strip().split()
    # Skip empty lines.
    if (len(line) == 0): return []
    # Remove a prefix if it is at the front of the line.
    for p in code.prefixes:
        if (line[0] == p):
            line = line[1:]
            break
    # Check for an empty line (this shouldn't happen).
    if (len(line) == 0):
        import warnings
        text = f"\nAn enexpected thing just happened when parsing.\n"+\
               f"After removing {class_name(code)} prefix '{p}', the line was empty.\n"
        warnings.warn(text+please_report_to)
    # Check for a match, if it matches complete instance initialization.
    elif (line[0] == code.type):
        parsed_code = code(list_of_lines, comments, parent)
        if (parsed_code.lines > 0): return [parsed_code]
        else:                       return []
    # No objects were found, return empty list of instances.
    return []

# Given a list of strings that represent lines of a file, determine if
# it is a recognizable declaration. If so, define Argument object(s),
# return the list of Argument(s).
def parse_argument(list_of_lines, comments, parent):
    line = list_of_lines[0].strip().split()
    success = False
    if (len(line) == 0): return []
    # Try out all possible declarations.
    from .real import Real
    from .integer import Integer
    from .logical import Logical
    from .character import Character
    from .complex import Complex
    from .type import TypeArgument
    for arg_type in [Real, Integer, Logical, Character, Complex, TypeArgument]:
        if (line[0] == arg_type.type):
            success = True
            break
    else: return []
    # If an argument type was identified, then finish parsing.
    double_colon = [i for i in range(len(line)-1)
                    if line[i] == line[i+1] == ":"]
    # If there is no "::", then variable names will immediately follow type(kind).
    # TODO: This might be a parsing error, or a FortranLanguageError.
    if (len(double_colon) > 1): raise(NotImplementedError)
    elif (len(double_colon) == 1):
        # There is a double colon in this line.
        double_colon = double_colon[0]
        base = line[:double_colon]
        tail = line[double_colon+2:]
    else:
        # If there is a KIND, then include that in the base.
        if ((len(line) > 2) and (line[1]) == '('):
            # Check to see of there is a paranthetical group after
            #   the arugment type (this would be for a KIND).
            kind, tail = pop_group(line[1:], open_with="(", close_with=")")
            base = [line[0]] + ["("] + kind + [")"]
        else:
            base = line[:1]
            tail = line[1:]
    # Check to make sure there are variable names after the TYPE.
    if (len(tail) == 0): raise(NotImplementedError)        
    # Now we are guaranteed to use this line to define an argument.
    list_of_lines.pop(0)
    # Get all names from the tail (with their dimensions, if given).
    names = [tail.pop(0)]
    while (len(tail) > 0):
        # First, check to see of there is a paranthetical group after
        #   the previous argument name (this would be for a size, like
        #   ARG(10,2), but it could have expresssions inside as well).
        group, tail = pop_group(tail, open_with="(", close_with=")")
        if (len(group) > 0): names[-1] += " ( "+" ".join(group)+" )"
        # If there is a "," it indicates more arguments follow.
        elif (tail[0] == ","): tail.pop(0)
        # If this is a PARAMETER assignment, there could be an "= value"
        #  after the name. There will only be other arguments after
        #  this one if there is a "," in the line. If no "," then stop.
        elif (tail[0] == "="):
            if (("," in tail) and ("=" in tail[1:])):
                tail = tail[tail.index(",")+1:]
            else: break
        # Finally, if it is not a comma, group, or value, it must be an argument name.
        else: names.append(tail.pop(0))
    # Cycle the names, converting all into proper Argument objects.
    args = []
    for n in names:
        args.append( arg_type(base[:]) )
        args[-1].parent = parent
        # Split the dimension out of the name, if it is there.
        if "(" in n:
            begin_dim = n.index("(")
            n, group = n[:begin_dim].strip(), n[begin_dim:].strip().split()
            group, _ = pop_group(group)
            dimension = [""]
            num_open = 0
            while (len(group) > 0):
                next_value = group.pop(0)
                if (next_value == ",") and (num_open == 0):
                    dimension.append("")
                else: dimension[-1] += next_value
                if (next_value == "("): num_open += 1
                if (next_value == ")"): num_open -= 1
            # Overwrite the dimension.
            args[-1].dimension = dimension
        # Overwrite the name.
        args[-1].name = n
    # Return the final list of declared arguments, and success.
    return args



from .argument import Argument

class Logical(Argument):
    type = "LOGICAL"
    c_types = {"4":"ctypes.c_int", "1":"ctypes.c_bool"}
    default_singleton = "0"

    # Add a warning about logical arrays.
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if (self.dimension is not None):
            # Change the C-type of this Logical to be an "int" for arrays.
            self.c_types = self.c_types.copy()
            self.c_types["4"] = "ctypes.c_int"
            # Show a warning to the user.
            from fmodpy.config import show_warnings
            if show_warnings:
                import warnings
                warnings.warn("Fortran LOGICAL arrays must be given as 32-bit integers.")




from fmodpy.parsing import FORT_TEXT_REPLACEMENTS, FORT_TEXT_FIXES

# Short functions.
#   identifying the strings before and after the last "."
def before_dot(name): return name[:len(name) - 1 - name[::-1].find(".")];
def after_dot(name): return name[min(-name[::-1].find("."),0):]
#   getting the class name of something, stripping module prefix
def class_name(cls): return str(type(cls)).split(".")[-1].split("'")[0]
#   identifying legal python module names
def legal_module_name(name):
    return (name.replace("_","")[0].isalpha() and
            name.replace("_","")[1:].isalnum() and
            (len(name) > 0))

# Shorten all strings to be 132 characters at maximum given a list of str.
def wrap_long_lines(list_of_lines, max_len=132):
    i = -1
    while (i+1 < len(list_of_lines)):
        i += 1
        # Get the line (strip off any comments for length checks).
        line = list_of_lines[i]
        if "!" in line: line = line[:line.index("!")]
        # Check the length of the (uncommented) line.
        if (len(line) > max_len-1):
            # Break up this line if it is too long.
            keep, rest = list_of_lines[i][:max_len-1], list_of_lines[i][max_len-1:]
            list_of_lines[i] = keep+"&"
            list_of_lines.insert(i+1, "&"+rest)
    return list_of_lines

# Function for efficiently performing a series of replacements on a line of text
def clean_text(text, replacements=FORT_TEXT_REPLACEMENTS, fixes=FORT_TEXT_FIXES):
    import re
    # Create a proper reg-exp dictionary of replacements
    replacements = {re.escape(k):v for (k,v) in replacements.items()}
    # Generate a regular expression pattern with that dictionary
    pattern = re.compile("|".join(replacements.keys()))
    # Perform the replacement using python reg-exp search
    cleaned_file = pattern.sub(lambda m: replacements[re.escape(m.group(0))], text)
    # Now repeat the above steps undoing any broken elements.
    fixes = {re.escape(k):v for (k,v) in fixes.items()}
    pattern = re.compile("|".join(fixes.keys()))
    fixed_file = pattern.sub(lambda m: fixes[re.escape(m.group(0))], cleaned_file)
    return fixed_file

# Read a fortran file, store it as single lines 
# (without leading and trailing whitespace) and return
def simplify_fortran_file(in_file, old_fortran=False):
    from fmodpy.parsing import ACCEPTABLE_LINE_STARTS, \
        LINE_STARTS_TO_REMOVE, IMMEDIATELY_EXCLUDE
        
    with open(in_file) as f:
        fort_file = []
        curr_line = ""
        for line in f.readlines():
            if (old_fortran) and (len(line) >0) and (line[0].upper() == "C"):
                line = "!" + line[1:]
            # Split out the comments from the line
            comment_start = line.find("!") if (line.find("!") != -1) else len(line)
            line, comment = line[:comment_start], line[comment_start:].strip()
            # Keep lines that are strictly comments (might be documentation)
            if len(line.strip()) == 0:
                fort_file.append(comment)
                continue
            # Make all fortran upper case
            line = line.upper()
            # Break the line by the colon character
            lines = line.split(";")
            for line in lines:
                if len(line.strip()) == 0: continue
                if (old_fortran):
                    line = line[:72]
                    # Line continuation character in column 5
                    if (len(line) > 5) and (line[5] in ["1","*"]):
                        line = "&" + line[6:]
                        # Retro-actively pop the previous line for continuation
                        if len(curr_line) == 0: curr_line = fort_file.pop(-1)
                    # Remove any numeric labels from the front of the line
                    line = line.strip().split()
                    if (len(line) > 0) and (line[0].isnumeric()): line = line[1:]
                    line = " ".join(line)
                # After processing old fortran, properly strip the line of whitespace
                line = line.strip()
                # Remove a leading ampersand if necessary
                if (line[0] == "&"):
                    line = line[1:]
                curr_line += line
                if ((len(line) > 0) and (line[-1] == "&")):
                    curr_line = curr_line[:-1]
                else:
                    # Process the line into a common format
                    #   (upper case, space separated, no commas)
                    clean_line = clean_text(curr_line)
                    line = [v.strip() for v in clean_line.split()]
                    # Only take lines that are acceptable
                    acceptable = (len(line) > 0) and (
                        ("!" in line[0]) or (line[0] in ACCEPTABLE_LINE_STARTS))
                    # Check for certain exclusions (like "PROGRAM").
                    if (len(line) > 0) and (line[0] in IMMEDIATELY_EXCLUDE):
                        from fmodpy.exceptions import FortranError
                        raise(FortranError(("A valid fortran python module cannot"+
                                            " contain '%s'.")%(line[0])))
                    # Store the line if it was acceptable.
                    if acceptable:
                        # Remove line starts that can be ignored safely.
                        while ((len(line) > 0) and (line[0] in LINE_STARTS_TO_REMOVE)):
                            line.pop(0)
                        # Add the line to the simplified Fortran file.
                        fort_file.append( " ".join(line) )
                    curr_line = ""
    return fort_file


# Given a list of strings, return the group of elements between
# <open_with> and <close_width>. This is initially designed to extract
# the elements within an open-close parenthesis while allowing for
# nested parenthetical groups. Returns two lists, one containing the
# group (if it starts at the beginning of <list_str>, else empty), the
# other containing the remainder of the list of strings.
def pop_group(list_str, open_with="(", close_with=")"):
    group = []
    # Get the first element of the string (open the group, if matched).
    if ((len(list_str) > 0) and (list_str[0] == open_with)):
        list_str.pop(0)
        num_open = 1
    else: num_open = 0
    # Search until the started group is closed.
    while (num_open > 0):
        # TOOD: Might need to raise parsing error when this happens.
        if (len(list_str) == 0): raise(NotImplementedError)
        next_value = list_str.pop(0)
        if   (next_value == open_with):  num_open += 1
        elif (next_value == close_with): num_open -= 1
        if (num_open != 0): group.append(next_value)
    # Return the captured portion and the remaining.
    return group, list_str
        



from .code import Code
from . import parse_code, parse_argument, parse_use, parse_type, \
    parse_implicit, parse_interface, parse_subroutine, parse_function

# ALLOCATABLE is not allowed to go from C -> Fortran. That means
#  INTENT(IN) is not allowed generally, and inside of a passed
#  PROCEDURE, INTENT(OUT) is not allowed.

# --------------------------------------------------------------------
class Subroutine(Code):
    type = "SUBROUTINE"
    prefixes = ["RECURSIVE"]
    can_contain = [(parse_use, "uses"),
                   (parse_implicit, "implicit_none"),
                   (parse_type, "types"),
                   (parse_interface, "interfaces"),
                   (parse_argument, "arguments"),
    ]
    will_ignore = [parse_subroutine, parse_function]
    bind_c = False # Set to 'True' in '.parse' if BIND(C) is found.

    # Produce a string representation of this Subroutine object as it
    # appears in the Fortran source (NOT as the wrapper will appear).
    def __str__(self):
        # Print out header line.
        out = f"{self.type} {self.name}({', '.join([a.name for a in self.arguments])}){'' if not self.bind_c else ' BIND(C)'}\n"
        # Add documentation.
        if (len(self.docs.strip()) > 0):
            doc_lines = self.docs.strip().split("\n")
            for line in doc_lines: out += f"  {line}\n"
        # Add used modules.
        for line in sorted(self.uses): out += "  "+line+"\n"
        for line in sorted(self.implicit_none): out += "  "+line+"\n"
        # Add types
        if (len(self.types) > 0): out += "\n"
        for t in self.types:
            for line in str(t).split("\n"):
                out += "  "+line+"\n"
        # Add interfaces.
        if (len(self.interfaces) > 0): out += "\n"
        for i in self.interfaces:
            for line in str(i).split("\n"):
                out += "  "+line+"\n"
        # Only add the space before arguments if types or
        # interfaces came before them.
        if ((max(len(self.interfaces),len(self.types)) > 0) and
            (len(self.arguments) > 0)): out += "\n"
        # Add arguments.
        for a in self.arguments: out += f"  {a}\n"
        # End the subroutine.
        out += f"END {self.type} {self.name}"
        return out

    # Given a list of lines (of a source Fortran file), parse out this
    # Subroutine (assuming the first line is the line declaring this
    # Subroutine).
    def parse(self, list_of_lines):
        self.lines += 1
        # Parse the name of this subroutine out of the argument list.
        declaration_line = list_of_lines.pop(0).strip().split()
        arg_start = declaration_line.index("(")
        arg_end = declaration_line.index(")")
        self.bind_c = "BIND(C)" in "".join(declaration_line[arg_end+1:])
        argument_names = declaration_line[arg_start+1:arg_end]
        argument_order = argument_names.copy()
        while "," in argument_names: argument_names.remove(",")
        self.name = declaration_line[:arg_start][-1]
        # ------- Default parsing operations -------
        super().parse(list_of_lines)
        # ------------------------------------------
        # Remove all declarations that are not arguments.
        to_remove = []
        for arg in self.arguments:
            if (arg.name in argument_names):
                argument_names.remove(arg.name)
            else:
                to_remove.append( arg )
        for arg in to_remove: self.arguments.remove(arg)
        # Check all interfaces for declarations of missing argument names.
        from .procedure import Procedure
        empty_interfaces = []
        for i in self.interfaces:
            # Check functions.
            to_remove = []
            for f in i.functions:
                if f.name in argument_names:
                    argument_names.remove(f.name)
                    self.arguments.append( Procedure([Procedure.type], parent=self) )
                    self.arguments[-1].name = f.name
                    self.arguments[-1].kind = "C_" + f.name
                    self.arguments[-1].intent = "IN"
                    self.arguments[-1].show_intent = False
                else:
                    to_remove.append(f)
            for f in to_remove: i.functions.remove(f)
            # Check subroutines.
            to_remove = []
            for f in i.subroutines:
                if f.name in argument_names:
                    argument_names.remove(f.name)
                    self.arguments.append( Procedure([Procedure.type], parent=self) )
                    self.arguments[-1].name = f.name
                    self.arguments[-1].kind = "C_" + f.name
                    self.arguments[-1].intent = "IN"
                    self.arguments[-1].show_intent = False
                else:
                    to_remove.append(f)
            for f in to_remove: i.subroutines.remove(f)
            # Mark this interface for removal, if it is empty.
            if ((len(i.functions) + len(i.subroutines)) == 0):
                empty_interfaces.append( i )
        # If there are any remaining undefined arguments..
        if (len(argument_names) > 0):
            from fmodpy.config import implicit_typing
            # Assign implicit INTEGER and REAL types to undefined arguments.
            if (implicit_typing):
                integer_letters = {'I', 'J', 'K', 'L', 'M', 'N'}
                for a in argument_names:
                    if (a[0] in integer_letters): line = f"INTEGER {a}"
                    else:                         line = f"REAL {a}"
                    arg = parse_argument([line], "", self)
                    assert (len(arg) == 1)
                    self.arguments.append( arg[0] )
            # Default behavior is to NOT support implicit typing.
            # Rather, assume that fmodpy has incorrectly parsed the
            # procedure.
            else:
                from fmodpy.exceptions import ParseError
                raise(ParseError(f"Finished parsing {self.type} {self.name}, but never declared {', '.join(argument_names)}."))
        # Remove empty interfaces.
        for i in empty_interfaces: self.interfaces.remove(i)
        # Sort the arguments in this subroutine according to their
        # position in the declared argument list.
        self.arguments.sort(key=lambda arg: argument_order.index(arg.name))


    # Return a list of lines that defines a Fortran -> C interface for
    # this routine. It will not contain the body code.
    def fortc_interface(self):
        # Define the interface to the C-function that will be called
        # from Fortran to evaluate externally provided routines.
        lines = []
        raise(NotImplementedError)

    # Return a list of lines that defines a Fortran -> C interface
    # that appropriate converts Fortran locals to C inputs.
    def fortc_subroutine(self):
        # Add the first line defining this FUNCTION / SUBROUTINE
        # Take all the normal inputs (given in the original INTERFACE)
        # Declare all arguments and local variables.
        # Translate Fortran arguments into C information.
        # Call C function (defined through abstract interface).
        # Translate C results back into Fortran arguments.
        raise(NotImplementedError)

    # Return a list of lines that defines a C -> Python interface
    # that allows Python code to be called from Fortran (through C).
    def cpy_function(self):
        # Translate C into NumPy values.
        # Call user-provided function, collect output.
        # Translate NumPy values back into C
        raise(NotImplementedError)

    # Generate Fortran wrapper code that can call this Subroutine's source code.
    def generate_fortran(self):
        # Generate the input signature.
        fortran_arguments = []
        for arg in self.arguments: fortran_arguments += arg.fort_input()
        # Regardless of if this is a FUNCTION or SUBROUTINE, we will always
        #  wrap with a SUBROUTINE to avoid RESULT related name conflicts.
        lines = ['',f"SUBROUTINE C_{self.name}({', '.join(fortran_arguments)}) BIND(C)"]
        # Add the "USE" line.
        lines += ["  "+l for l in self.uses]
        if any((a.allocatable and a._is_output()) for a in self.arguments):
            lines += ["  USE ISO_FORTRAN_ENV, ONLY: INT64"]
        # Check if this is in a module.
        in_module = (self.parent is not None) and (self.parent.type == "MODULE")
        if (in_module): lines += [f"  USE {self.parent.name}, ONLY: {self.name}"]
        # Add any type definitions from parents that are used here.
        known_types = {t.name for t in self.types}
        for l in self.uses:
            if (":" not in l): continue
            known_types |= {n.strip() for n in l.split(":")[-1].split(",")}
        needed_types = set()
        for a in self.arguments:
            if ((a.type == "TYPE") and (a.kind not in known_types)):
                needed_types.add(a.kind)
        # Assume that all needed types are in the parent.
        if (len(needed_types) > 0):
            parent_types = set()
            for t in self.parent.types:
                if (t.name in needed_types):
                    parent_types.add(t.name)
                    needed_types.remove(t.name)
            lines += [f"  USE {self.parent.name}, ONLY: {', '.join(sorted(parent_types))}"]
        # Enforce no implicit typing (within this code).
        lines += [f"  IMPLICIT NONE"]
        # Add all type definitions.
        for t in self.types:
            lines += ["  " + l for l in t.fort_declare()]
        # Add all argument declarations.
        for arg in self.arguments:
            lines += ["  " + l for l in arg.fort_declare()]
        lines += ['']
        # If this is not inside of a module, need to define an interface.
        if ((not in_module) or (len(self.interfaces) > 0)):
            # Add the "INTERFACE" line for the source subroutine.
            lines += ["  INTERFACE"]
            # Need to define the interface to the C function that
            #   will be called by the internal Fortran wrapper.
            for i in self.interfaces:
                for f in i.functions:   lines += ["    "+l for l in f.fortc_interface()]
                for f in i.subroutines: lines += ["    "+l for l in f.fortc_interface()]
            # Need to define the interface to the actual function
            # being called (if this is not part of a module).
            if (not in_module):
                # Force this routine to have IMPLICIT NONE for interface definition.
                self.implicit_none = ["IMPLICIT NONE"]
                # Add the interface definition of this routine.
                lines += ["    "+l for l in str(self).split("\n")]
            lines += ["  END INTERFACE",'']
        # Add all argument preparation code.
        for arg in self.arguments:
            lines += ['  '+l for l in arg.fort_prepare()]
        # Add the call line.
        optional_args = [a for a in self.arguments if a.optional]
        mandatory_args = [a for a in self.arguments if not a.optional]
        im_a_function = hasattr(self, "result")
        # Set the beginning of the line that calls the source routine.
        if im_a_function:
            for arg in mandatory_args:
                if (im_a_function and (arg.name == self.result)):
                    call_start = f"{arg.fort_call_name()} ="
                    mandatory_args.remove(arg)
                    break
        else: call_start = "CALL"
        # Generate the calling code recursively when there are optionals.
        if (len(optional_args) > 0):
            # Warn the user if there are a lot of optional arguments.
            if (len(optional_args) > 10):
                import warnings
                warnings.warn("\nfmodpy [Subroutine.generate_fortran]:\n"+
                             f" '{self.name}' has {len(optional_args)} OPTIONAL arguments, which\n"+
                             f"  means the wrapper will have 2^{len(optional_args)} = {2**len(optional_args)}\n"+
                              "  IF .. ELSE .. ENDIF statements to handle all combinations.\n"+
                              "  Consider reducing the number of OPTIONAL arguments, or be prepared\n"+
                              "  for a large wrapper file and long compilation times.")
            # If there are any optionals, assign all arguments
            # literally in the call to Fortran (because the ordering
            # could be off after accounting for presence).
            always_assigned = [f"{a.name}={a.fort_call_name()}" for a in mandatory_args]
            # Define a recursive routine to write out all cases for the presence
            # of the optional variables using nested 'IF PRESENT()' conditions.
            def make_call_with_optionals(present, missing, optionals,
                                         indent="  ", lines=lines):
                if (len(optionals) > 0):
                    a = optionals[0]
                    lines += [f"{indent}IF ({a.fort_present_name()}) THEN"]
                    make_call_with_optionals(present+[a], missing, optionals[1:], indent+"  ")
                    lines += [f"{indent}ELSE"]
                    make_call_with_optionals(present, missing+[a], optionals[1:], indent+"  ")
                    lines += [f"{indent}END IF"]
                else:
                    # Get all assigned names.
                    assigned = [f"{a.name}={a.fort_call_name()}" for a in present]
                    call_args = ", ".join(always_assigned+assigned)
                    # Add the call line.
                    lines += [f"{indent}{call_start} {self.name}({call_args})"]
                    # Produce lines that come after the call that are
                    # specific to the presence of the optional arguments.
                    for a in present: lines += [indent+l for l in a.fort_after(present=True)]
                    for a in missing: lines += [indent+l for l in a.fort_after(present=False)]
            # Use the recursive call-generating function.
            make_call_with_optionals([], [], optional_args)
        else:
            call_args = ', '.join([a.fort_call_name() for a in mandatory_args])
            lines += [f"  {call_start} {self.name}({call_args})"]

        # Put the "RESULT" back into the list of mandatory arguments.
        if im_a_function:
            mandatory_args += [a for a in self.arguments if (a.name == self.result)]

        # Add all the argument post-processing code (for mandatory arguments).
        for arg in mandatory_args:
            fort_after = arg.fort_after()
            if (len(fort_after) > 0): fort_after = ['']+fort_after
            lines += ['  '+l for l in fort_after]

        # Add a declaration of all abstract interfaces (if there are any)
        # for calling back into C from inside of Fortran.
        if (len(self.interfaces) > 0):
            lines += ['', 'CONTAINS']
            for i in self.interfaces:
                # TODO: Need to define the interface to the C function that
                #       will be called by the internal Fortran wrapper.
                for f in i.functions: lines += ["  "+l for l in f.fortc_subroutine()]
                for f in i.subroutines: lines += ["  "+l for l in f.fortc_subroutine()]
            lines += ['']

        # Add the END line.
        lines += [f"END SUBROUTINE C_{self.name}",'']
        return lines

    # Generate Python code that accesses this Subroutine.
    def generate_python(self, type_blocks=None):
        from fmodpy.config import fmodpy_print as print
        lines = [ '',
                  "# ----------------------------------------------",
                 f"# Wrapper for the Fortran subroutine {self.name}",
                  '']
        py_name = self.name.lower()
        # Add the Python-callable function.
        py_input = []
        # If this is in a module, then the first attribute will be "self".
        in_module = (self.parent is not None) and (self.parent.type == "MODULE")
        if in_module: py_input.append("self")
        # If there is no parent to capture the types, make them in the subroutine.
        type_lines = []
        if (type_blocks is None):
            # Cycle through and create all type declarations where necessary.
            for t in self.types:
                type_lines += t.py_declare()
        # Otherwise, add the type blocks to the parent set to be processed.
        else:
            for t in self.types:
                type_blocks.add("\n".join(t.py_declare()))
        # Cycle args (make sure the ones that are optional are listed last).
        for arg in sorted(self.arguments, key=lambda a: int(a._is_optional())):
            py_input += arg.py_input()
            # Add the python type declaration blocks for complex if appropriate.
            # TODO: `complex` type definition should be stored in `self.types`
            #       instead of being stored in the 'py_type' attribute for
            #       an argument. The Argument should not define the type.
            arg_py_type = arg.py_type
            if (arg_py_type is not None):
                # Add lines for the complex type before this routine.
                if (type_blocks is None):
                    lines += [""]
                    lines += ["    "+l for l in arg_py_type.split("\n")]
                # Add lines for this complex type to the parent.
                else:
                    type_blocks.add(arg_py_type)
        # Declare the function and add the documentation.
        lines += [f"def {py_name}({', '.join(py_input)}):",
                  f"    '''{self.docs}'''"]
        # Import any types that are imported in the corresponding fortran.
        known_types = {t.name for t in self.types}
        get_type_kind = lambda arg: arg.kind.split("(")[0].strip()
        needed_types = {get_type_kind(a) for a in self.arguments if a.type == "TYPE"
                        and get_type_kind(a) not in known_types}
        #   extract types directly from referenced modules that should be wrapped
        for l in self.uses:
            type_names = l.split()
            type_names.pop(0)
            class_name = type_names.pop(0).lower()
            if (':' in type_names):
                type_names = type_names[type_names.index(':')+1:]
            to_pop = set()
            for t in sorted(needed_types):
                if t in type_names:
                    to_pop.add(t)
                    lines += [f"    {t} = {class_name}.{t}"]
            needed_types -= to_pop
        #   extract types directly from the possible chain of code's parents
        p = self.parent
        while (p is not None):
            temp_py_name = p.name.lower()
            for t in getattr(p, "types", []):
                needed_types -= {getattr(t, "name", "")}
                lines += [f"    {t.name} = {temp_py_name}.{t.name}"]
            p = getattr(p, "parent", None)
        # Check for errors (needed types must be explicitly imported).
        if (len(needed_types) > 0):
            from fmodpy.exceptions import NotSupportedError
            raise(NotSupportedError(
                f"\nArguments for routine '{self.name}' were derived TYPE, but the type was not explicitly imported." +
                f"\nAll derived types must either be defined in the subroutine where they are used, or explicitly imported from a wrapped module."
            ))
        # Declare any internal types if they could not be put in the parent.
        if (len(type_lines) > 0):
            lines += ['    '+l for l in type_lines]
        # Add the declaration lines.
        py_declare = []
        for arg in self.arguments:
            arg_declare = arg.py_declare()
            if (len(arg_declare) > 0):
                py_declare += ['',f'# Setting up "{arg.name.lower()}"'] + arg_declare
        lines += ['    '+l for l in py_declare]
        # Add call line.
        py_call = []
        for arg in self.arguments: py_call += arg.py_call()
        lines += ['',"    # Call C-accessible Fortran wrapper.",
                    f"    clib.c_{py_name}({', '.join(py_call)})", '']
        # Add post-processing line.
        py_after = []
        for arg in self.arguments:
            arg_after = arg.py_after()
            if (len(arg_after) > 0):
                arg_after = [f'# Post-processing "{arg.name.lower()}"']+ arg_after +['']
            py_after += arg_after
        lines += ["    "+l for l in py_after]
        # Add return line.
        py_return = []
        for arg in self.arguments: py_return += arg.py_return()
        lines += [ "    # Return final results, 'INTENT(OUT)' arguments only.",
                  f"    return {', '.join(py_return)}"]
        return lines




from .code import Code
from . import parse_use, parse_implicit, parse_public, parse_private, \
    parse_argument, parse_interface, parse_type, parse_subroutine, \
    parse_function
from .subroutine import Subroutine
from .function import Function


# --------------------------------------------------------------------
class Module(Code):
    type = "MODULE"
    status = "PUBLIC"
    can_contain = [(parse_use, "uses"),
                   (parse_implicit, "implicit_none"),
                   (parse_public, "public"),
                   (parse_private, "private"),
                   (parse_type, "types"),
                   (parse_interface, "interfaces"),
                   (parse_argument, "arguments"),
                   (parse_subroutine, "subroutines"),
                   (parse_function, "functions"),
    ]


    # Produce a string representation of this code object.
    def __str__(self):
        out = f"{self.type} {self.name}\n"
        # Add documentation.
        if (len(self.docs.strip()) > 0):
            doc_lines = self.docs.strip().split("\n")
            for line in doc_lines: out += f"  {line}\n"
        # Add used modules.
        for line in sorted(self.uses): out += f"  {line}\n"
        for line in sorted(self.implicit_none): out += f"  {line}\n"
        # Add types.
        for t in self.types:
            for line in str(t).split("\n"):
                out += f"  {line}\n"
        # Add interfaces.
        if (len(self.interfaces) > 0): out += "\n"
        for i in self.interfaces:
            for line in str(i).split("\n"):
                out += f"  {line}\n"
        # Only add the space before arguments if types or
        # interfaces came before them.
        if ((max(len(self.interfaces),len(self.types)) > 0) and
            (len(self.arguments) > 0)): out += "\n"
        # Add arguments.
        for a in self.arguments:
            out += f"  {a}\n"
        # Add subroutines.
        for obj in self.subroutines:
            out += "\n"
            for line in str(obj).split("\n"):
                out += f"  {line}\n"
        # Add functions.
        for obj in self.functions:
            out += "\n"
            for line in str(obj).split("\n"):
                out += f"  {line}\n"
        # End the module.
        out += f"END {self.type} {self.name}\n"
        return out

    # Generate Fortran wrapper code that can call this Module's source code.
    def generate_fortran(self):
        # Generate the input signature.
        fortran_arguments = []
        for arg in self.arguments: fortran_arguments += arg.fort_input()
        lines = ['',f'{self.type} C_{self.name}']
        # Add the "USE" line.
        lines += self.uses
        if any((a.allocatable and a._is_output()) for a in self.arguments):
            lines += ['  USE ISO_FORTRAN_ENV, ONLY: INT64']
        # Enforce no implicit typing (within this code).
        lines += [f'  IMPLICIT NONE']
        lines += ['']

        # Declare all interfaces.
        for i in self.interfaces:
            if (len(i.subroutines) + len(i.functions) > 0):
                lines += ['']
                lines += ['  '+l for l in str(i).split("\n")]

        lines += ['']
        lines += ['CONTAINS']
        lines += ['']

        # Define all "getter" and "setter" subroutines for accessing
        # internal module (PUBLIC) attributes.
        for arg in self.arguments:
            lines += ['',f'  ! Getter and setter for {arg.name}.']
            lines += ['  '+l for l in arg.fort_getter()]
            lines += ['  '+l for l in arg.fort_setter()]

        # Declare all subroutines and functions.
        for code in self.subroutines + self.functions:
            lines += ['']
            lines += ['  ' + l for l in code.generate_fortran()]

        # Add the END line.
        lines += [f"END {self.type} C_{self.name}",'']

        return lines

    # Generate Python code.
    def generate_python(self, *args):
        # Add access points for the contained subroutines and functions.
        lines = [ '',
                 f'class {self.name.lower()}:',
                 f"    '''{self.docs}'''"]
        # Generate internal types.
        for t in self.types:
            lines += [ "    "+l for l in t.py_declare()]
        # Generate getter and setter for arguments.
        for arg in self.arguments:
            py_name = arg.name.lower()
            lines += ['']
            lines += [f"    # Declare '{py_name}'"]
            # Declare the "getter" for this module variable.
            lines += [ "    "+l for l in arg.py_getter() ]
            # Declare the "setter" for this module variable.
            lines += [ "    "+l for l in arg.py_setter() ]
            # Declare this as a property that has a getter and a setter.
            lines += [ "    "+l for l in arg.py_property() ]
        # Generate static methods for all internal routines.
        for code in self.subroutines + self.functions:
            lines += ['']
            lines += ['    '+l for l in code.generate_python(*args)]
        # Replace the class definition with a single instance of the class.
        lines += [ "",
                  f"{self.name.lower()} = {self.name.lower()}()"]
        return lines

    # Given a list of lines (of a source Fortran file), parse out this
    # module (assuming the first line is the first line *inside* of
    # this module).
    def parse(self, list_of_lines):
        self.lines += 1
        line = list_of_lines.pop(0).strip().split()
        # Remove any comments on the line that may exist.
        if ("!" in line): line = line[:line.index("!")]
        # Catch a potential parsing error.
        if (len(line) <= 1):
            from fmodpy.exceptions import ParseError
            raise(ParseError(f"Expected 'MODULE <NAME>', but the line did not contain the name of module.\n  {list_of_lines[0]}"))
        # Get the name of this module.
        self.name = line[1]
        # ------- Default parsing operations -------
        super().parse(list_of_lines)
        # ------------------------------------------
        # Get the set of things declared private.
        private = set(self.private)
        # Remove any instances of things that are declared private.
        if (len(private) > 0):
            for attr in ("arguments","interfaces","types","subroutines","functions"):
                codes = getattr(self,attr)
                to_remove = [i for (i,c) in enumerate(codes) if (c.name in private)]
                for i in reversed(to_remove): codes.pop(i)
                if (len(to_remove) > 0):
                    from fmodpy.config import fmodpy_print as print
                    print(f"  removed {len(to_remove)} from '{attr}' declared PRIVATE..")
        # Remove any things not declared public if this MODULE is private.
        if (self.status == "PRIVATE"):
            public = set(self.public)
            for attr in ("interfaces","types","subroutines","functions"):
                codes = getattr(self,attr)
                to_remove = [i for (i,c) in enumerate(codes) if (c.name not in public)]
                for i in reversed(to_remove): codes.pop(i)
                if (len(to_remove) > 0):
                    from fmodpy.config import fmodpy_print as print
                    print(f"  removed {len(to_remove)} from '{attr}' because this MODULE is PRIVATE..")
        # Make sure all "arguments" don't show intent.
        for a in self.arguments: a.show_intent = False



# - make immutable INTENT(OUT) objects not part of input.
# - make test cases (and necessary code) for:
#     real
#        inputs
#        outputs
#        array inputs
#        array outputs (known size)
#        array outputs (unknown size)
#        optional inputs
# 
#        optional array inputs
#        optional array outputs (known size)
#        optional array outputs (unknown size)
#        allocatable outputs
#        optional allocatable outputs
#     double precision
#        ^^
#     integer
#        ^^
#     logical
#        ^^
#     character
#        ^^
#     type
#        ^^
#     undefined inputs
#     procedure inputs
#     POINTER attribute
#     modules with interfaces
#     modules with internal variables (define getter / setter functions)
#     modules with custom type declarations
#     type EXTENDS (copy the other type, add more definitions)
#     fixed format Fortran inputs (Fortran 77)
# 
#     
# "Module" should be put inside a Python class. Then I can define
# proper "getter" and "setter" behaviors for internal attributes.
# 


def run_tests():
    from fmodpy.config import load_config
    load_config(verbose=False, verbose_module=False, wrap=True, rebuild=True)

    # Import all of the tests.
    from fmodpy.test.character import test as test_character
    from fmodpy.test.real32 import test as test_real32
    from fmodpy.test.real64 import test as test_real64
    from fmodpy.test.double_precision import test as test_double_precision
    from fmodpy.test.int32 import test as test_int32
    from fmodpy.test.int64 import test as test_int64
    from fmodpy.test.logical import test as test_logical
    from fmodpy.test.module import test as test_module
    from fmodpy.test.complex64 import test as test_complex64
    from fmodpy.test.complex128 import test as test_complex128
    from fmodpy.test.type import test as test_type
    from fmodpy.test.misc import test as test_misc

    # Run all of the tests.
    test_character()
    test_real32()
    test_real64()
    test_double_precision()
    test_int32()
    test_int64()
    test_logical()
    test_module()
    test_complex64()
    test_complex128()
    test_type()
    test_misc()


if __name__ == "__main__":
    # /Users/thomaslux/Library/Python/3.7/bin/pprofile 
    # 
    # Activate a code coverage module.
    # import coverage, os
    # this_dir = os.path.dirname(os.path.abspath(__file__))
    # module_dir = os.path.dirname(this_dir)
    # omitted = ([this_dir] + [os.path.join(this_dir,f) for f in os.listdir(this_dir)]
    #            + [os.path.join(this_dir,f)+"*" for f in os.listdir(this_dir)
    #               if os.path.isdir(os.path.join(this_dir,f))])
    # print("Omitted:")
    # print(omitted)
    # cov = coverage.Coverage(source=[module_dir], omit=omitted)
    # cov.start()

    # Run the tests on `fmodpy`.
    run_tests()

    exit()

    print()
    print("Generating code coverage report..", flush=True)
    # Save data from the coverage tracking (for report).
    cov.stop()
    cov.save()
    print("  (will host local webserver for 10 minutes or until killed)", flush=True)
    # Create a temporary directory for holding test results.
    from tempfile import TemporaryDirectory
    temp_dir = TemporaryDirectory()
    results_dir = temp_dir.name
    cov.html_report(directory=results_dir)
    # Open the results file.
    import webbrowser
    webbrowser.open("file://"+os.path.join(results_dir, "index.html"))
    # Wait for load, then delete the temporary directory.
    #   (user might decide to kill this process early).
    import time
    time.sleep(60*10) # <- Default to 10 minutes of waiting.
    temp_dir.cleanup()
    del temp_dir



# Making this module accessible by being called directly from command line.
import os, sys, traceback

# Import all of the fmodpy module
from .fmodpy import __doc__, fimport, configure

if len(sys.argv) < 2:
    import fmodpy
    help(fmodpy)
    exit()
else:
    file_path = os.path.abspath(sys.argv[1])

# Pretty error handling when this file is executed directly     
def custom_excepthook(exc_type, value, tb):
    l = ''.join(traceback.format_exception(exc_type, value, tb))
    print(l)
sys.excepthook = custom_excepthook

# Read command line arguments after the path to the source file.
command_line_args = {}
for arg in sys.argv[2:]:
    if ("=" not in arg):
        from fmodpy.exceptions import UnrecognizedConfiguration
        raise(UnrecognizedConfiguration(
            f"Command line argument {str([arg])[1:-1]} must be formatted"+
            " as '<setting>=<value>', but no '=' found."))
    first_equals = arg.index('=')
    setting, value = arg[:first_equals], arg[first_equals+1:]
    command_line_args[setting] = value

# Call "fimport" providing the given arguments as configurations.
fimport(file_path, **command_line_args)




# Standardized testing interface.
def test():
    import os
    import shutil
    dir_name = os.path.dirname(os.path.abspath(__file__))
    test_name = os.path.basename(dir_name)
    build_dir = os.path.join(dir_name, f"fmodpy_{test_name}")
    import fmodpy
    # ---------------------------------------------------------------
    # Begin specific testing code.
    fort_file = os.path.join(dir_name, f"subroutine_with_type.f90")
    print(f"  {os.path.basename(fort_file)}..", end=" ", flush=True)
    fort = fmodpy.fimport(fort_file, build_dir=build_dir,
                          output_dir=dir_name, rebuild=True,
    )
    assert (fort.one_up(1) == fort.FANCY(shirt=2.0, pants=3)), "Failed!"
    print("passed", flush=True)
    shutil.rmtree(os.path.join(dir_name,fort_file[:-4]))
    # End specific testing code.
    # ---------------------------------------------------------------


if __name__ == "__main__":
    test()



# Custom errors that can be raised during the fmodpy process

class BadKeywordArgument(Exception): pass
class BuildAndLinkError(Exception): pass
class CompileError(Exception): pass
class FortranError(Exception): pass
class IllegalConfiguration(Exception): pass
class LinkError(Exception): pass
class NameError(Exception): pass
class NotAllowedPath(Exception): pass
class NotSupportedError(Exception): pass
class ParseError(Exception): pass
class SizeError(Exception): pass
class UnrecognizedConfiguration(Exception): pass



# Set up some package-related things.
import os
DIRECTORY = os.path.dirname(os.path.abspath(__file__))
ABOUT_DIR = os.path.join(DIRECTORY, "about")
with open(os.path.join(ABOUT_DIR,"version.txt")) as f:
    __version__ = f.read().strip()
del f

# Steal the documentation from .fmodpy, instead of duplicating here.
from .fmodpy import __doc__

# Import the main configuration (needed to initialize the configuration).
import fmodpy.config

# Import the main features of this package.
from .fmodpy import fimport, configure, load_symbol

# Set "__all__" so that "from fmodpy import *" returns expected stuff.
__all__ = ["fimport", "configure", "load_symbol"]




# Standardized testing interface.
def test():
    import os
    dir_name = os.path.dirname(os.path.abspath(__file__))
    test_name = os.path.basename(dir_name)
    fort_file = os.path.join(dir_name, f"test_{test_name}.f03")
    build_dir = os.path.join(dir_name, f"fmodpy_{test_name}")
    print(f"  {test_name}..", end=" ", flush=True)
    import fmodpy
    fort = fmodpy.fimport(fort_file, build_dir=build_dir,
                          output_dir=dir_name, verbose=False, wrap=True,
                          rebuild=True, show_warnings=False)
    # ---------------------------------------------------------------
    # Begin specific testing code.

    import ctypes
    import numpy as np
    a = np.array(list(map(ord,['1', '0', '1', '0', '1', '0', '1', '0'])),
                 dtype=ctypes.c_char, order='F')
    out = a.copy()
    # Check for successful copy.
    result = fort.test_simple_character(a, b=out, c=ord('1'))
    assert(all(result.view('uint8') == a.view('uint8')))
    # Check for successful overwrite.
    result = fort.test_simple_character(a, b=out, c=ord('0'))
    assert(all(result.view('uint8') == [1, 2, 0, 1, 2, 0, 1, 2]))

    # End specific testing code.
    # ---------------------------------------------------------------
    print("passed", flush=True)
    import shutil
    shutil.rmtree(os.path.join(dir_name,f"test_{test_name}"))

if __name__ == "__main__":
    test()



from .argument import Argument

class Character(Argument):
    type = "CHARACTER"
    c_types = {"1":"ctypes.c_char"}
    default_singleton = "0"

    # Add a warning about logical arrays.
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if (self.dimension is not None):
            # TODO: Change the C-type of this character be a "uint8" for arrays.
            # self.c_types = self.c_types.copy()
            # self.c_types["1"] = "ctypes.c_char" # Needs to be uint8 in numpy.
            # Show a warning to the user.
            from fmodpy.config import show_warnings
            if show_warnings:
                import warnings
                warnings.warn("Fortran CHARACTER arrays are not supported yet. Raise an issue with an example to encourage development.")



from .subroutine import Subroutine

# --------------------------------------------------------------------
class Function(Subroutine):
    type = "FUNCTION"
    result = None

    # Produce a string representation of this Subroutine object.
    def __str__(self):
        # Print out header line.
        arg_names = [a.name for a in self.arguments if (a.name != self.result)]
        out = f"{self.type} {self.name}({', '.join(arg_names)})"
        if (self.result != self.name): out += f" RESULT({self.result})\n"
        else:                          out += "\n"
        # Use SUBROUTINE string method for the rest (strip SUBROUTINE).
        out += "\n".join(super().__str__().split("\n")[1:])
        return out

    # Given a list of lines (of a source Fortran file), parse out this
    # Function (assuming the first line is the first line *inside* of
    # this Function).
    def parse(self, list_of_lines):
        # Pre-fetch the "result" and modify the first line so that it
        # can be parsed correctly by the Subroutine.parse function.
        declaration_line = list_of_lines[0].strip().split()
        arg_start = declaration_line.index("(")
        arg_end = declaration_line.index(")")
        argument_names = declaration_line[arg_start+1:arg_end]
        while "," in argument_names: argument_names.remove(",")
        name = declaration_line[:arg_start][-1]
        rest_of_line = declaration_line[arg_end+1:]
        # Get the result from the end of the line.
        if ("RESULT" in rest_of_line):
            if ("(" not in rest_of_line) or (")" not in rest_of_line):
                from fmodpy.exceptions import ParseError
                raise(ParseError(f"Found 'RESULT' but no '(<name>)' on line.\n{list_of_lines[0].strip()}"))
            self.result = ''.join(rest_of_line[rest_of_line.index("(")+1:
                                               rest_of_line.index(")")])
        else: self.result = name
        argument_names.append( self.result )
        list_of_lines[0] = f"SUBROUTINE {name} ( {' , '.join(argument_names)} )"
        # ------------------------------------------------------------
        # Use Subroutine parse code.
        super().parse(list_of_lines)
        # ------------------------------------------------------------
        # Find the result in the identified arguments, make sure its
        # intent is not declared (it will make Fortran compilers angry).
        for arg in self.arguments:
            if (arg.name == self.result):
                arg.intent = "OUT"
                arg.show_intent = False
                break
        else:
            from fmodpy.config import fmodpy_print as print
            print()
            print("Did not find output argument in parsed list..")
            print(self.result)
            print()
            raise(NotImplementedError)




from .code import Code
from . import parse_subroutine, parse_function, parse_module, parse_type

# --------------------------------------------------------------------
# Class for holding all relevant information in a Fortran file.
class Fortran(Code):
    docs = ""
    type = "FILE"
    needs_end = False
    allowed_unknown = True
    can_contain = [(parse_subroutine, "subroutines"),
                   (parse_function, "functions"),
                   (parse_module, "modules"),
                   (parse_type, "types"),
    ]

    # Generate a string describing this file.
    def __str__(self):
        out = f"{self.type} {self.name}"
        for (_, name) in self.can_contain:
            if (len(getattr(self, name)) > 0): out += "\n"
            for obj in getattr(self,name):
                for line in str(obj).split("\n"):
                    out += "  "+line+"\n"
        return out


    # Create Fortran Subroutines with "BIND(C)" that are
    # accessible from C, translate arguments to Fortran, call the
    # Fortran code, and translates arguments back (if necessary).
    def generate_fortran(self):
        lines = ["! This automatically generated Fortran wrapper file allows codes",
                 "! written in Fortran to be called directly from C and translates all",
                 "! C-style arguments into expected Fortran-style arguments (with",
                 "! assumed size, local type declarations, etc.).",
                 ""]
        lines += super().generate_fortran()
        # Return the list of lines for this object (add an extra newline).
        from . import wrap_long_lines
        # Return the final Fortran file as a string.
        return "\n".join(wrap_long_lines(lines+['']))

    # Generate a string representation of the python code for this wrapper.
    def generate_python(self):
        # Define the header to this file. Some necessary variables are
        # left to be filled by formatting, namely:
        #    f_compiler          -- string
        #    shared_object_name  -- string
        #    f_compiler_args     -- string
        #    dependencies        -- list of strings
        #    module_name         -- string
        lines = ["'''This Python code is an automatically generated wrapper",
                 "for Fortran code made by 'fmodpy'. The original documentation",
                 "for the Fortran source code follows.",
                 "",
                 self.docs,
                 "'''",
                 '',
                 'import os',
                 'import ctypes',
                 'import platform',
                 'import numpy',
                 '',
                 '# --------------------------------------------------------------------',
                 '#               CONFIGURATION',
                 '# ', 
                 '_verbose = {verbose_module}',
                 '_fort_compiler = "{f_compiler}"',
                 '_shared_object_name = "{shared_object_name}." + platform.machine() + ".so"',
                 '_this_directory = os.path.dirname(os.path.abspath(__file__))',
                 '_path_to_lib = os.path.join(_this_directory, _shared_object_name)',
                 '_compile_options = {f_compiler_args}',
                 '_ordered_dependencies = {dependencies}',
                 '_symbol_files = {symbol_files}'
                 '# ',
                 '# --------------------------------------------------------------------',
                 '#               AUTO-COMPILING',
                 '#',
                 '# Try to import the prerequisite symbols for the compiled code.',
                 'for _ in _symbol_files:',
                 '    _ = ctypes.CDLL(os.path.join(_this_directory, _), mode=ctypes.RTLD_GLOBAL)',
                 '# Try to import the existing object. If that fails, recompile and then try.',
                 'try:',
                 '    clib = ctypes.CDLL(_path_to_lib)',
                 'except:',
                 '    # Remove the shared object if it exists, because it is faulty.',
                 '    if os.path.exists(_shared_object_name):',
                 '        os.remove(_shared_object_name)',
                 '    # Compile a new shared object.',
                 '    _command = " ".join([_fort_compiler] + _compile_options + ["-o", _shared_object_name] + _ordered_dependencies)',
                 '    if _verbose:',
                 '        print("Running system command with arguments")',
                 '        print("  ", _command)',
                 '    # Run the compilation command.',
                 '    import subprocess',
                 '    subprocess.run(_command, shell=True, cwd=_this_directory)',
                 '    # Import the shared object file as a C library with ctypes.',
                 '    clib = ctypes.CDLL(_path_to_lib)',
                 '# --------------------------------------------------------------------',
                 '']
        # Add all the Python lines for the children of this Code.
        lines += super().generate_python()
        # Return the Python file as a string.
        return "\n".join(lines + [''])



from . import pop_group

# --------------------------------------------------------------------
# Base class for all "Argument" objects that contain information about
# the pieces of data that need to be passed between Python and Fortran.
class Argument:
    parent = None # Container of this argument (Module, Subroutine, etc.)
    type = "" # argument type (REAL, INTEGER, etc.)
    name = "" # name of this argument in code
    size = "" # size in bytes (uses compiled code to evaluate)
    kind = "" # KIND setting
    intent = "INOUT" # INTENT setting
    show_intent = True # if True then include INTENT in __str__
    kind_prefix = "KIND=" # String that precedes kind declaration in Fortran
    allocatable = False # May be (re)allocated at runtime.
    optional = False # May be excluded from input.
    save = False # Save value between executions.
    value = False # Pass by value, not by reference.
    pointer = False # Pass by pointer, not by reference.
    parameter = False # Whether this argument is a parameter (set at compile time)
    type_len = False # Whether this argument has the LEN parameter (in a derived TYPE)
    type_kind = False # Whether this argument has the KIND parameter (in a derived TYPE)
    dimension = None # If array, this list describes the shape and size.
    c_types = {} # The C-types used to declare this argument (key is self.size)
    c_types_arrays = {} # The C-types used to declare this argument as arrays.
    default_singleton = "1" # The default value assigned to a singleton.

    # Properties are used to add wrapper logic around retrieving types.
    @property
    def c_type(self):
        if (self.size not in self.c_types):
            raise(NotImplementedError(f"\n\nUnrecognized size '{self.size}' for argument '{self.name}', no known corresponding C type."))
        return self.c_types[self.size]
    @property
    def c_type_array(self):
        return self.c_types_arrays.get(self.size, self.c_type)
    @property
    def py_type(self): return None

    # ----------------------------------------------------------------
    #                  Generating Fortran Wrapper

    # Return the name of this variable that should be used locally,
    # safe for use with RESULT values from FUNCTIONs that do not
    # define a separate name for results.
    # 
    # WARNING: This function returns a string not a list, because the
    #          size checking will have an exact 1-1 mapping of arguments.
    def function_safe_name(self):
        name = self.name
        if (self.parent is not None) and (self.parent.name == self.name):
            name += "_RESULT"
        return name

    # Return the string name of this variable for "PRESENT" checking.
    # WARNING: This is only used by "OPTIONAL" arguments.
    def fort_present_name(self): return self.name + "_PRESENT"

    # WARNING: This function returns a string not a list, because the
    #          Fortran call will have an exact 1-1 mapping of arguments.
    def fort_call_name(self):
        name = self.name
        if (self.parent is not None) and (self.parent.name == self.name):
            name += "_RESULT"
        elif (self.allocatable):
            name += "_LOCAL"
        return name

    # Names of arguments that will be given to the Fortran wrapper.
    def fort_input(self):
        names = []
        # Add a boolean "_PRESENT" for this variable if optional.
        if (self.optional):
            names.append(self.name+"_PRESENT")
        # Add extra arguments for the dimension sizes.
        if (self.dimension is not None):
            for i in range(len(self.dimension)):
                names.append(f"{self.name}_DIM_{i+1}")
        # Append the actual name of this variable.
        names.append(self.name)
        # Add an identifier to this variable if it has the name of a function.
        if (self.parent is not None) and (self.parent.name == self.name):
            names[-1] += "_RESULT"
        return names

    # Lines of Fortran code necessary to declare variables for this argument.
    def fort_declare(self):
        lines = []
        if (self.optional):
            lines.append(f"LOGICAL, INTENT(IN) :: {self.name}_PRESENT")
        # If this is an allocatable, the sizes have to be given as output.
        size_type = "IN" if not self.allocatable else "OUT"
        # Add extra arguments for the dimension sizes.
        if (self.dimension is not None):
            og_dimension = self.dimension[:]
            for i in range(len(self.dimension)):
                lines.append(f"INTEGER(KIND=SELECTED_INT_KIND(18)), INTENT({size_type}) :: {self.name}_DIM_{i+1}")
                # If this is allocatable, the dimension cannot be assumed on input.
                if self.allocatable: self.dimension[i] = ":"
                else: self.dimension[i] = f"{self.name}_DIM_{i+1}"
        # Add a local variable that will have the "ALLOCATABLE" metadata.
        if self.allocatable:
            temp_arg = self.copy()
            temp_arg.save = True
            temp_arg.name += "_LOCAL"
            temp_arg.show_intent = False
            temp_arg.optional= False
            lines.append(str(temp_arg))
        # Append actual name of this variable (with dimension
        # values filled to be the names of the integer inputs 
        # and OPTIONAL / ALLOCATABLE turned off).
        temp_arg = self.copy()
        if (self.parent is not None) and (self.parent.name == self.name):
            temp_arg.name += "_RESULT"
        temp_arg.optional = False
        temp_arg.allocatable = False
        # If this is allocatable, it will be returned as an INT64 memory address.
        if self.allocatable:
            temp_arg.dimension = None
            temp_arg.intent = "OUT"
            temp_arg.show_intent = True
            temp_arg.type = "INTEGER"
            temp_arg.kind = "INT64"
        # Add the actual definition of this argument.
        lines.append(str(temp_arg))
        # Reset the dimension, if it had one beforehand.
        if (self.dimension is not None): self.dimension = og_dimension
        # Return the lines of declaration.
        return lines

    # Lines of Fortran code that must be executed before the call.
    def fort_prepare(self): return []

    # Lines of Fortran code that must be executed after the call.
    def fort_after(self, present=True):
        lines = []
        is_output = ("OUT" in self.intent) or (self.intent == "")
        # Copy over the size of the output allocatable array if it was not known.
        if (is_output and self.allocatable):
            if (self.dimension is None): raise NotImplementedError
            for i in range(1,len(self.dimension)+1):
                if present: size = f"SIZE({self.function_safe_name()}_LOCAL,{i})"
                else:       size = "0"
                lines.append(f"{self.function_safe_name()}_DIM_{i} = {size}")
            # Copy the allocatable address only when it is present.
            if present:
                first_pos = ",".join(["1"]*len(self.dimension))
                lines.append(f"{self.function_safe_name()} = LOC({self.name}_LOCAL({first_pos}))")
        # Return the list of lines.
        return lines

    # ----------------------------------------------------------------
    #                  Generating Python Wrapper

    # The names of arguments as seen in the Python function declaration.
    def py_input(self):
        # If this is not allowed as input, do not include it.
        if (not self._allowed_input()): return []
        # This object is now either mutable or partially an input.
        elif self._is_optional(): return [self.name.lower()+"=None"]
        else: return [self.name.lower()]
        
    # The lines of Python code to execute before the Fortran call.
    def py_declare(self):
        lines = []
        # This is where output arrays whose dimensions are known can
        # be initialized. Everything else behaves normally.
        py_name = self.name.lower()
        # If this is literally an optional, a boolean is required.
        if self.optional:
            lines.append(f"{py_name}_present = ctypes.c_bool(True)")
        # If inputs can be given, and they're optional, checks are needed.
        if (self._allowed_input() and self._is_optional()):
            lines.append(f"if ({py_name} is None):")
            # If this is an optional argument, make it "not present".
            if self.optional:
                lines.append(f"    {py_name}_present = ctypes.c_bool(False)")
            # If this is a singleton, set its default value.
            if (self.dimension is None):
                lines.append(f"    {py_name} = {self.c_type}()")
                lines.append(f"else:")
                lines.append(f"    {py_name} = {self.c_type}({py_name})")
            # This has a dimension AND is allocatable.
            elif (self.allocatable):
                if (self.optional):
                    default_shape = ','.join('0'*len(self.dimension))
                    lines.append(f"    {py_name} = numpy.zeros(shape=({default_shape}), dtype={self.c_type_array}, order='F')")
                    lines.append(f"elif (type({py_name}) == bool) and ({py_name}):")
                else:
                    lines.append(f"    {py_name} = ctypes.c_void_p()")
                default_shape = ','.join('1'*len(self.dimension))
                # If the size is known, then we can initialize this optional array to pass it in.
                if (self._known_size()):
                    default_shape = ', '.join(self._default_size()).lower()
                # Create lines for initializing the default values.
                if (self._is_input() or self.optional):
                    lines.append(f"    {py_name} = numpy.zeros(shape=({default_shape}), dtype={self.c_type_array}, order='F')")
                else:
                    lines.append(f"    {py_name} = ctypes.c_void_p()")
            # This has a dimension, but is NOT allocatable.
            else:
                default_shape = ','.join('1'*len(self.dimension))
                # If this is optional (not present shape is always 1).
                if (self.optional):
                    lines.append(f"    {py_name} = numpy.zeros(shape=({default_shape}), dtype={self.c_type_array}, order='F')")
                    lines.append(f"elif (type({py_name}) == bool) and ({py_name}):")
                # If the size is known, then we can initialize this optional array to pass it in.
                if (self._known_size()):
                    default_shape = ', '.join(self._default_size()).lower()
                # Create lines for initializing the default values.
                lines.append(f"    {py_name} = numpy.zeros(shape=({default_shape}), dtype={self.c_type_array}, order='F')")
        # Convert appropriate array-inputs into Fortran compatible arrays.
        if ((self.dimension is not None) and (self._allowed_input())):
            # Check for Fortran-continuity and data type of array inputs.
            p, s = "", "" # <- prefix, spaces
            if self._is_optional(): p, s = f"el", "  "
            lines += [f"{p}if ((not issubclass(type({py_name}), numpy.ndarray)) or",
                      f"{s}    (not numpy.asarray({py_name}).flags.f_contiguous) or",
                      f"{s}    (not ({py_name}.dtype == numpy.dtype({self.c_type_array})))):",
                       "    import warnings",
                      f"    warnings.warn(\"The provided argument '{py_name}' was not an f_contiguous NumPy array of type '{self.c_type_array}' (or equivalent). Automatically converting (probably creating a full copy).\")",
                      f"    {py_name} = numpy.asarray({py_name}, dtype={self.c_type_array}, order='F')",]
        # If this is an output-only allocatable, declare a local pointer.
        elif ((not self.optional) and self._is_output() and self.allocatable):
            lines.append(f"{py_name} = ctypes.c_void_p()")
        # Otherwise this an output immutable type, declare it locally.
        elif (not self._allowed_input()):
            lines.append(f"{py_name} = {self.c_type}()")
        # This is a singleton input, convert to appropraite C type.
        elif (self._is_input() and (self.dimension is None)):
            lines.append(f"if (type({py_name}) is not {self.c_type}): {py_name} = {self.c_type}({py_name})")
        # Define the dimension size variables if appropriate.
        if (self.dimension is not None):
            # If this object is allowed as input..
            if (self._allowed_input()):
                # This is optional, so we need to check what to do with dimensions.
                if (self.optional):
                    lines.append(f"if ({py_name}_present):")
                    for i in range(len(self.dimension)):
                        lines.append(f"    {py_name}_dim_{i+1} = ctypes.c_long({py_name}.shape[{i}])")
                    lines.append("else:")
                    for i in range(len(self.dimension)):
                        lines.append(f"    {py_name}_dim_{i+1} = ctypes.c_long()")
                    if self.allocatable:
                        lines.append(f"{py_name} = ctypes.c_void_p({py_name}.ctypes.data)")
                # This is not optional, so it is declared at this point, get the dimensions.
                else:
                    for i in range(len(self.dimension)):
                        lines.append(f"{py_name}_dim_{i+1} = ctypes.c_long({py_name}.shape[{i}])")
            # This array is output, initialize dimension memory locations.
            else:
                for i in range(len(self.dimension)):
                    lines.append(f"{py_name}_dim_{i+1} = ctypes.c_long()")
        # Return lines of declaration code.
        return lines

    # The names of arguments passed to the "BIND(C)" Fortran code.
    def py_call(self):
        names = []
        py_name = self.name.lower()
        # Add a boolean "_PRESENT" for this variable if optional.
        if (self.optional): names.append("ctypes.byref("+py_name+"_present)")
        # Add extra arguments for the dimension sizes.
        if (self.dimension is not None):
            for i in range(len(self.dimension)):
                names.append(f"ctypes.byref({py_name}_dim_{i+1})")
        # Append the actual name of this variable.
        names.append(py_name)
        # If this is a memory view, then access the first element
        # (start of block of memory).
        if ((self.dimension is not None) and (not self.allocatable)):
            names[-1] = "ctypes.c_void_p(" + names[-1] + ".ctypes.data)"
        else:
            names[-1] = "ctypes.byref(" + names[-1] + ")"
        # Return the list of names.
        return names

    # The lines of Python code that must be executed after the Fortran call.
    def py_after(self):
        lines = []
        if self.allocatable and self._is_output():
            py_name = self.name.lower()
            # This must be an array argument and it must be 'INTENT(OUT)'.
            if (self.dimension is None): raise(NotImplementedError)
            # Get the pointer to the first index.
            local_dims = [f"{py_name}_dim_{i+1}.value" for i in range(len(self.dimension))]
            # Compute the size of the (flattened) array.
            lines += [f"{py_name}_size = ({') * ('.join(local_dims)})"]
            shape = ','.join(local_dims[::-1])
            # Make a line that checks if the array should be None.
            if self.optional:
                check_line = [f"if ({py_name}_present) and ({py_name}_size > 0):"]
            else:
                check_line = [f"if ({py_name}_size > 0):"]
            lines += check_line
            # Get flat array pointer and reshape by the dimensions
            # reversed, then transpose (f_contiguous).
            lines += [f"    {py_name} = numpy.array(ctypes.cast({py_name}, ctypes.POINTER({self.c_type}*{py_name}_size)).contents, copy=False)"]
            # If the array type does not match the singleton type, use a view.
            if (self.c_type != self.c_type_array):
                lines[-1] += f".view({self.c_type_array})"
            # If this is a tensor, then reshape it (from C style, row major) to Fortran style.
            if (len(self.dimension) > 1):
                lines += [f"    {py_name} = {py_name}.reshape({shape}).T"]
            # Otherwise, (if not present) this output is None.
            lines += [f"elif ({py_name}_size == 0):",
                      f"    {py_name} = numpy.zeros(shape=({shape}), dtype={self.c_type_array}, order='F')",
                      f"else:",
                      f"    {py_name} = None"]
        # Return all lines.
        return lines

    # The name of the argument in the final "return" line to Python.
    def py_return(self):
        names = []
        if (self._is_output()):
            # Transform array memory views back into NumPy arrays.
            is_pointer = (self._is_output() and self.allocatable)
            # Add this argument to the output.
            py_name = self.name.lower()
            value = py_name
            # Retrieve the Python "value" of the C object if appropriate.
            if (self.dimension is None) and (self.type != "TYPE"):
                value += ".value"
            # Return None for missing optional returns.
            if self.optional: names.append(f"({value} if {py_name}_present else None)")
            else:             names.append(f"{value}")
        # Return the processed name.
        return names

    # ----------------------------------------------------------------
    #                    Module Attribute Access

    # Define a 'getter' function in Python for a public module attribute.
    def py_getter(self):
        py_name = self.name.lower()
        call_args = []
        lines = [f'def get_{py_name}(self):']
        # If this is not in a MODULE, then that was unexpected.
        if ((self.parent == None) or (self.parent.type != "MODULE")):
            raise(NotImplementedError)
        # If this module attribute is allocatable, we might need to return None.
        if (self.allocatable):
            call_args.append(f'ctypes.byref({py_name}_allocated)')
            lines.append(f'    {py_name}_allocated = ctypes.c_bool(False)')
        # Define the dimension size variables if appropriate.
        if (self.dimension is not None):
            for i in range(len(self.dimension)):
                lines.append(f'    {py_name}_dim_{i+1} = ctypes.c_long()')
                call_args.append(f'ctypes.byref({py_name}_dim_{i+1})')
                # Define the actual attribute holder itself.
        if (self.allocatable or (self.dimension is not None)):
            lines += [f'    {py_name} = ctypes.c_void_p()']
        else:
            lines += [f'    {py_name} = {self.c_type}()']
        call_args.append(f'ctypes.byref({py_name})')
        # Make the call to the Fortran wrapped function for getting.
        module_name = self.parent.name.lower()
        lines += [f'    clib.{module_name}_get_{py_name}({", ".join(call_args)})']
        # If this allocatable is not allocated, return "None".
        if (self.allocatable):
            lines.append(f'    if (not {py_name}_allocated.value): return None')
        # If this is an array, retreive its value from a pointer.
        if (self.dimension is not None):
            # Get the pointer to the first index.
            local_dims = [f"{py_name}_dim_{i+1}.value" for i in range(len(self.dimension))]
            # Compute the size of the (flattened) array.
            lines.append(f"    {py_name}_size = ({') * ('.join(local_dims)})")
            # Get flat array pointer and reshape by the dimensions
            # reversed, then transpose (f_contiguous).
            lines.append(f"    if ({py_name}_size > 0):")
            lines.append(f"        {py_name} = numpy.array(ctypes.cast({py_name}, ctypes.POINTER({self.c_type}*{py_name}_size)).contents, copy=False)")
            # If the array type does not match the singleton type, use a view.
            if (self.c_type != self.c_type_array):
                lines[-1] += f".view({self.c_type_array})"
            lines.append(f"    else:")
            lines.append(f"        {py_name} = numpy.zeros((0,), dtype={self.c_type}, order='F')")
            # If this is a tensor, then reshape it to be a tensor again.
            if (len(self.dimension) > 1):
                shape = ','.join(local_dims[::-1])
                lines += [f"    {py_name} = {py_name}.reshape({shape}).T"]
            # Return the value.
            lines += [f'    return {py_name}']
        else:
            lines += [f'    return {py_name}.value']
        # Return the lines that compose this "getter".
        return lines

    # Define a 'setter' function in Python for a public module attribute.
    def py_setter(self):
        py_name = self.name.lower()
        lines = [f'def set_{py_name}(self, {py_name}):']
        if self.parameter:
            lines += ["    raise(NotImplementedError('Module attributes with PARAMETER status cannot be set.'))"]
            return lines
        call_args = []
        # If this is not in a MODULE, then that was unexpected.
        if ((self.parent == None) or (self.parent.type != "MODULE")):
            raise(NotImplementedError)
        # Get the size if there is a dimension.
        if (self.dimension is not None):
            lines += [f"    if ((not issubclass(type({py_name}), numpy.ndarray)) or",
                      f"        (not numpy.asarray({py_name}).flags.f_contiguous) or",
                      f"        (not ({py_name}.dtype == numpy.dtype({self.c_type})))):",
                       "        import warnings",
                      f"        warnings.warn(\"The provided argument '{py_name}' was not an f_contiguous NumPy array of type '{self.c_type}' (or equivalent). Automatically converting (probably creating a full copy).\")",
                      f"        {py_name} = numpy.asarray({py_name}, dtype={self.c_type}, order='F')",]
            # Store all the sizes.
            for i in range(len(self.dimension)):
                dim_name = f'{py_name}_dim_{i+1}'
                lines.append(f'    {dim_name} = ctypes.c_long({py_name}.shape[{i}])')
                call_args.append(f'ctypes.byref({dim_name})')
            # Call passing in all arguments (might include sizes).
            call_args.append(f'ctypes.c_void_p({py_name}.ctypes.data)')
        else:
            # Initialize correct c_type version of value.
            lines.append(f'    {py_name} = {self.c_type}({py_name})')
            # Call passing in all arguments (might include sizes).
            call_args.append(f'ctypes.byref({py_name})')
        module_name = self.parent.name.lower()
        lines += [f'    clib.{module_name}_set_{py_name}({", ".join(call_args)})']
        return lines

    # Define the 'property' referencing the correct getter and setter.
    def py_property(self):
        py_name = self.name.lower()
        module_name = self.parent.name.lower()
        getter_name = f"get_{py_name}"
        setter_name = f"set_{py_name}"
        return [f'{py_name} = property({getter_name}, {setter_name})']

    # Define the FORTRAN BIND(C) subroutine for retreiving the value
    # of this module attribute.
    def fort_getter(self):
        # Add extra arguments for the dimension sizes.
        args = []
        decs = [f"USE {self.parent.name}, ONLY: {self.name}"] # declarations
        lines = []
        # Create a local copy.
        temp = self.copy()
        temp.name += "_LOCAL"
        # Get allocatable present input.
        if (self.allocatable):
            args.append(f'{self.name}_ALLOCATED')
            decs.append(f'LOGICAL, INTENT(OUT) :: {self.name}_ALLOCATED')
            lines += [f'{self.name}_ALLOCATED = ALLOCATED({self.name})',
                      f'IF (.NOT. {self.name}_ALLOCATED) RETURN']

        # Get all dimension size inputs.
        if (self.dimension is not None):
            temp.type = "INTEGER"
            temp.kind = "INT64"
            temp.dimension = None
            temp.allocatable = False
            decs.insert(0,"USE ISO_FORTRAN_ENV, ONLY: INT64")
            for i in range(len(self.dimension)):
                args.append(f"{self.name}_DIM_{i+1}")
                decs.append(f"INTEGER(KIND=INT64), INTENT(OUT) :: {self.name}_DIM_{i+1}")
                lines.append(f"{self.name}_DIM_{i+1} = SIZE({self.name}, {i+1}, KIND=INT64)")
        # Disable "parameter" status for the local copy.
        if self.parameter: temp.parameter = False
        # Add argument for the actual variable.
        args.append(f"{temp.name}")
        decs += [str(temp)]
        # Insert the declarations before the lines.
        lines = decs + lines
        # Do the assignment.
        if (self.dimension is None):
            lines.append(f"{temp.name} = {self.name}")
        else:
            first_pos = ",".join(["1"]*len(self.dimension))
            lines.append(f"{temp.name} = LOC({self.name}({first_pos}))")
        # Add indentation to all lints.
        lines = ["  "+l for l in lines]
        # Add the subroutine line (with all arguments).
        lines.insert(0, f"SUBROUTINE {self.parent.name}_GET_{self.name}({', '.join(args)}) BIND(C)")
        # Add the end of the subroutine declaration line.
        lines.append(f"END SUBROUTINE {self.parent.name}_GET_{self.name}")
        return lines

    # Define the FORTRAN BIND(C) subroutine for setting the value
    # of this module attribute.
    def fort_setter(self):
        # If this is a parameter, it cannot be set.
        if (self.parameter): return []
        # Add extra arguments for the dimension sizes.
        args = []
        lines = []
        # Get all dimension size inputs.
        temp = self.copy()
        temp.name += "_LOCAL"
        temp.dimension = []
        temp.allocatable = False
        if (self.dimension is not None):
            for i in range(len(self.dimension)):
                args.append(f"{self.name}_DIM_{i+1}")
                lines.append(f"INTEGER(KIND=SELECTED_INT_KIND(18)), INTENT(IN) :: {self.name}_DIM_{i+1}")
                temp.dimension.append(args[-1])
        # Add argument for the actual variable.
        args.append(temp.name)
        # Insert the declaration for the local (input / set) value.
        lines += [str(temp)]
        # Get allocatable present input.
        if (self.allocatable):
            lines += [f'IF (ALLOCATED({self.name})) THEN',
                      f'  DEALLOCATE({self.name})',
                       'END IF']
            shape_str = ','.join([f'1:{name}' for name in temp.dimension])
            lines += [f'ALLOCATE({self.name}({shape_str}))']
        lines += [f"{self.name} = {temp.name}"]
        # Add indentation to all lints.
        lines = ["  "+l for l in lines]
        # Add the subroutine line (with all arguments).
        lines.insert(0, f"SUBROUTINE {self.parent.name}_SET_{self.name}({', '.join(args)}) BIND(C)")
        lines.insert(1, f"  USE {self.parent.name}, ONLY: {self.name}")
        # Add the end of the subroutine declaration line.
        lines.append(f"END SUBROUTINE {self.parent.name}_SET_{self.name}")
        return lines


    # ----------------------------------------------------------------
    #                    Wrapper Utilities

    # Return True if this variable has INTENT(IN).
    def _is_input(self): return ("IN" in self.intent)

    # Return True if this variable has INTENT(OUT).
    def _is_output(self): return ("OUT" in self.intent)

    # Check to see if this argument has known size.
    def _known_size(self): return (len(self._dimension_args()) == 0)

    # Return a list of integers that are the indices of dimensions
    # (1-indexed) that require user input definition.
    def _dimension_args(self):
        if (self.dimension is None): return []
        return [i+1 for (i,d) in enumerate(self.dimension) if d in {":","*"}]

    # Check to see if this argument is allowed as input.
    def _allowed_input(self):
        # This has "IN" in the intent.
        if self._is_input(): return True
        # This variable is strictly OUT.
        elif (self.optional): return True
        elif (self.dimension is None): return False
        elif (self.allocatable): return False
        # Could be given pre-allocated array.
        return True

    # Return True if this argument is optional (for Python users).
    def _is_optional(self):
        # This is explicitly declared optional.
        if (self.optional): return True
        # This has "IN" in the intent.
        elif (self._is_input()): return False
        # This is not allowed to be optional, because it has no input.
        elif (not self._allowed_input()): return False
        # This is an INTENT(OUT) argument.
        #   Size is known already.
        elif (self._known_size()): return True
        #   Size will be known after.
        elif (self.allocatable): return True
        #   Need to know the size.
        else: return False
        
    # Compute the default size of this Argument (assuming that
    # len(self._dimension_args()) == 0), in terms of Python, converted
    # from the assumed size of the object in Fortran.
    def _default_size(self):
        if (len(self._dimension_args()) != 0): raise(NotImplementedError)
        import re # <- use regular expressions to capture SIZE(...) syntax.
        sizes = []
        for size in self.dimension:
            size = size.lower().replace(" ","")
            # Replace all occurrences of "SIZE(<argument>)"
            #   with a Python NumPy equivalent array syntax.
            while "size(" in size:
                start_index = size.index("size(") + len("size(")
                before_call = size[:start_index-len("size(")]
                parens = 1
                for index in range(start_index, len(size)):
                    if size[index] == "(": parens += 1
                    elif size[index] == ")": parens -= 1
                    if parens <= 0: break
                argument = size[start_index:index]
                after_call = size[index+len(")"):]
                # Now "argument" contains whatever was inside the parenthesis
                #  and passed into the `SIZE(...)` function in Fortran.
                argument = argument.strip()
                # If this is just a reference to a variable name...
                if (re.match(r"^\w+$", argument) is not None):
                    func_replacement = f"{argument}.size"
                # If this is formatted as "<variable name>, <integer>", get that dimension..
                elif (re.match(r"^\w+,\s*\d+$", argument) is not None):
                    name = argument[:argument.index(",")]
                    dim = argument[argument.index(",")+1:]
                    dim = str(int(dim) - 1)
                    func_replacement = f"{name}.shape[{dim}]"
                # If this is formatted as "<variable name>(<slice>[,<slice>])", translate the slice(s)..
                elif (re.match(r"^\w+[(]\s*(:|[\w]+)\s*(,\s*(:|[\w]+)\s*)*[)]$", argument) is not None):
                    name = argument[:argument.index("(")]
                    slices = argument[argument.index("(")+1:argument.index(")")].split(",")
                    for i in range(len(slices)):
                        if (slices[i].strip() == ":"): continue
                        try:
                            slices[i] = str(int(slices[i])-1)
                        except ValueError:
                            slices[i] += "-1"
                    func_replacement = f"{name}[{','.join(slices)}].size"
                # Otherwise, the contents of `SIZE(...)` cannot be parsed successfully.
                else:
                    raise(NotImplementedError("The contents of 'SIZE' could not be parsed successfully. If this is valid Fortran syntax, consider raising an issue with a minimum necessary example from this code.\n\n  "+size))
                # Replace float division with integer division in python.
                size = before_call + func_replacement + after_call
            # Append this NumPy size.
            sizes.append(size)
        # fix any module attribute names to have "self." before them
        if (hasattr(self.parent, "parent") and
            hasattr(self.parent.parent, "type") and
            (self.parent.parent.type == "MODULE")):
            # create a temporary function that says whether or
            #  not a character could belong to a python name
            is_not_py_char = lambda c: not (c.isalpha() or c.isdigit() or (c == '_'))
            # cycle the internal attribute of the parent module that
            #  might be accessible here
            for arg in self.parent.parent.arguments:
                name = arg.name.lower()
                # cycle all size entries to check their contents
                for i in range(len(sizes)):
                    # if the module attribute is used and it is not just
                    #  a substring match, then make the replacement
                    s = sizes[i]
                    if (name in s):
                        j = s.index(name)
                        if ((j == 0) or is_not_py_char(s[j-1])):
                            sizes[i] = s.replace(name, "self."+name)
        # Return the list of sizes.
        return sizes

    # ----------------------------------------------------------------
    #                      Generic Methods

    # Generate a deep copy of this argument and return it.
    def copy(self):
        arg = type(self)([self.type])
        for attr in dir(self):
            # Skip hidden attributes.
            if (attr[:1] == "_"): continue
            # TODO: Come up with better general way to handle these special cases.
            if (attr in {"c_type", "c_type_array", "py_type"}): continue
            value = getattr(self, attr)
            # Skip executable attributes.
            if (hasattr(value, "__call__")): continue
            # If this attribute has a ".copy" function, then use it.
            if (hasattr(value, "copy") and hasattr(value.copy, "__call__")):
                value = value.copy()
            # Set the attribute in the new argument.
            setattr(arg, attr, value)
        return arg
    

    # Default initialization, process standard Fortran specifications.
    # Expects to be given list of strings that comes from a declaration
    # line, but with the ":: name1, ..." stripped off the end.
    def __init__(self, line, parent=None):
        if (len(line) == 0): raise(NotImplementedError)
        # Make sure this line matches the expected type.
        if (line.pop(0) != self.type): raise(NotImplementedError)
        # Set the parent.
        self.parent = parent
        # If the line is empty now, then we're done (use defaults).
        if (len(line) == 0): return
        # Parse the remainder of the declaration line.
        # 
        # Get the KIND declaration, if it was given.
        group, line = pop_group(line)
        if (len(group) > 0):
            if (tuple(group[:2]) == ("KIND","=")):
                group = group[2:]
            self.kind = " ".join(group)
        # The rest can come in any order, so loop over possibilities.
        while (len(line) > 0):
            # Skip commas.
            if (line[0] == ","):
                line.pop(0)
                continue
            # The following happen if the ":: names" has not been stripped.
            if (line[0] == ":"): raise(NotImplementedError)
            # Read ALLOCATABLE
            if (line[0] == "ALLOCATABLE"):
                line.pop(0)
                self.allocatable = True
            # Read DIMENSION
            elif (line[0] == "DIMENSION"):
                group, line = pop_group(line[1:])
                if (len(group) == 0): raise(NotImplementedError)
                # Break up dimensions by commas, allow nested parentheses.
                self.dimension = [""]
                num_open = 0
                while (len(group) > 0):
                    next_value = group.pop(0)
                    if (next_value == ",") and (num_open == 0):
                        self.dimension.append("")
                    else: self.dimension[-1] += next_value
                    if (next_value == "("): num_open += 1
                    if (next_value == ")"): num_open -= 1
                # Done processing dimension.
            # Read INTENT
            elif (line[0] == "INTENT"):
                group, line = pop_group(line[1:])
                if (len(group) == 0): raise(NotImplementedError)
                self.intent = "".join(group)
            # Read OPTIONAL
            elif (line[0] == "OPTIONAL"):
                line.pop(0)
                self.optional = True
            # Read SAVE
            elif (line[0] == "SAVE"):
                line.pop(0)
                self.save = True
            # Read PARAMETER (discard this)
            elif (line[0] == "PARAMETER"):
                line.pop(0)
                self.parameter = True
            # Read EXTERNAL (discard this)
            elif (line[0] == "EXTERNAL"): 
                line.pop(0)
                import warnings
                warnings.warn("fmodpy.parsing.argument: Ignoring 'EXTERNAL' status of argument.")
            # Read LEN attribute (for inside of TYPE declarations)
            elif (line[0] == "LEN"):
                line.pop(0)
                self.type_len = True
            # Read KIND attribute (for inside of TYPE declarations)
            elif (line[0] == "KIND"):
                line.pop(0)
                self.type_kind = True
            # Otherwise, this is an unrecognized argument.
            else:
                raise(NotImplementedError(f"\n\nUnrecognized part of Argument '{line[0]}'.\n"))

    # Print the Fortran string declaration of this argument.
    def __str__(self):
        out = f"{self.type}"
        if (len(self.kind) > 0): out += f"({self.kind_prefix}{self.kind})"
        if (self.show_intent and (len(self.intent) > 0)): out += f", INTENT({self.intent})"
        if (self.type_len): out += ", LEN"
        if (self.type_kind): out += ", KIND"
        if (self.parameter): out += ", PARAMETER"
        if (self.optional): out += f", OPTIONAL"
        if (self.allocatable): out += f", ALLOCATABLE"
        if (self.save): out += f", SAVE"
        if ((self.dimension is not None) and
            (len(self.dimension) > 0)):
            out += f", DIMENSION({','.join(self.dimension)})"
        if (len(self.name) > 0): out += f" :: {self.name}"
        # Return the final string.
        return out



# Custom error (in case user does not have setuptools)
class DependencyError(Exception): pass

# Try to import setuptools (if it fails, the user needs that package)
try: 
    from setuptools import setup, find_packages
except:
    raise(DependencyError("Missing python package 'setuptools'.\n  python3 -m pip install --user setuptools"))

import os
# Go to the "about" directory in the package directory
PACKAGE_NAME = "fmodpy"
DEFAULT_DIRECTORY = os.path.join(os.path.dirname(os.path.abspath(__file__)),PACKAGE_NAME,"about")
# Convenience function for reading information files
def read(f_name, dir_name=DEFAULT_DIRECTORY, processed=True):
    text = []
    # Check for directory existence.
    if (not os.path.exists(dir_name)):
        print(f"ERROR: No directory found '{dir_name}'")
        while (not os.path.exists(dir_name)):
            dir_name = os.path.dirname(dir_name)
        print(f"       only found {os.listdir(dir_name)}.")
        return [""] if processed else ""
    # Check for path existence.
    path = os.path.join(dir_name, f_name)
    if (not os.path.exists(path)):
        print(f"ERROR: No path found '{path}'")
        print(f"       only found {os.listdir(dir_name)}.")
        return [""] if processed else ""
    # Process the contents and return.
    with open(path) as f:
        if processed:
            for line in f:
                line = line.strip()
                if (len(line) > 0) and (line[0] != "%"):
                    text.append(line)
        else:
            text = f.read()
    # Return the text.
    return text

if __name__ == "__main__":
    #      Read in the package description files     
    # ===============================================
    package = PACKAGE_NAME
    version =read("version.txt")[0]
    description = read("description.txt")[0]
    requirements = read("requirements.txt")
    keywords = read("keywords.txt")
    classifiers = read("classifiers.txt")
    name, email, git_username = read("author.txt")

    setup(
        author=name,
        author_email=email,
        name=package,
        packages=find_packages(exclude=[]),
        include_package_data=True,
        install_requires=requirements,
        version=version,
        url = f"https://github.com/{git_username}/{package}",
        download_url = f"https://github.com/{git_username}/{package}/archive/{version}.tar.gz",
        description = description,
        keywords = keywords,
        python_requires = ">=3.6",
        license="MIT",
        classifiers=classifiers
    )

    # Check to see if the user has a fortran compiler. Create a
    # warning if there is no Fortran compiler installed.




# Standardized testing interface.
def test():
    import os
    dir_name = os.path.dirname(os.path.abspath(__file__))
    test_name = os.path.basename(dir_name)
    fort_file = os.path.join(dir_name, f"test_{test_name}.f03")
    build_dir = os.path.join(dir_name, f"fmodpy_{test_name}")
    print(f"  {test_name}..", end=" ", flush=True)
    import fmodpy
    fort = fmodpy.fimport(fort_file, build_dir=build_dir,
                          output_dir=dir_name)
    # ---------------------------------------------------------------
    # Begin specific testing code.

    import numpy as np
    n = 10
    array_in = np.asarray(np.arange(n), dtype="float32", order='F')
    array_out = np.zeros(n//2 - 1, dtype="float32", order='F')
    out = fort.test_standard(7, array_in, array_out)
    # Test the standard functionality.
    assert(out[0] == 8)
    assert(all(out[1] == [0,1,2,3]))
    assert(all(out[2] == [1,2,3,4]))
    assert(np.all(out[3] == [[2,3,4,5],
                             [3,4,5,6],
                             [4,5,6,7]]))
    assert(out[4] == None)
    # Test the extended functionality.
    out = fort.test_extended(4)
    assert(out[0] == None)
    assert(out[1] == None)
    assert(all(out[2] == [4, 3, 2, 1]))
    # WARNING: In the current code, the memory associated with out[2]
    #          will be freed by Fortran on subsequent calls to the
    #          function "test_extended". Copy for object permanance.
    out2 = fort.test_extended(10, known_opt_array_out=True)
    assert(all(out2[0] == [1,2,3]))
    assert(out2[1] == None)
    assert(all(out2[2] == list(reversed(range(11)[1:]))))
    # WARNING: Similar to before, the memory at out2[0] and out2[2]
    #          will be freed by Fortran on subsequent calls to
    #          "test_extended". Copies should be made for permanence.
    out3 = fort.test_extended(6, opt_alloc_array_out=True)
    assert(out3[0] == None)
    assert(all(out3[1] == [3,2,1]))
    assert(all(out3[2] == list(reversed(range(7)[1:]))))

    # End specific testing code.
    # ---------------------------------------------------------------
    print("passed", flush=True)
    import shutil
    shutil.rmtree(os.path.join(dir_name,f"test_{test_name}"))

if __name__ == "__main__":
    test()




# --------------------------------------------------------------------
#                              fmodpy
# 
#       an automatic fortran wrapper (and importer) for python.
# 
# This program is designed to processes a source fortran file and wrap
# all contained code into a python module.
# 
#   fimport -- Given a path to a Fortran source file,
#              wrap it and return a Python module.
# 
# For configuration and future usage of `fmodpy` on a single machine,
# use the following function.
# 
#   configure -- Three behaviors related to `fmodpy` configuration:
#                  no arguments -> Print current configuration and return.
#                  str, str, ... -> Remove these configurations from global config file.
#                  key=value, ... -> Assign these configurations in global config file.
# 
# --------------------------------------------------------------------

# Overwrite the "print" function to add "FMODPY" specific printouts.
from fmodpy.config import fmodpy_print as print
from fmodpy.config import libraries

# Function for automating the entire fortran<->python wrapping
#   process. This is the top-level function built for standard usage.
#   This function automatically checks the modification times of the
#   output module as well as the source fortran in order to only
#   re-compile when necessary. Automatically wraps fortran, makes
#   symbolic links to all files in the source directory as potential
#   dependencies for building, generates a wrapper in Fortran, 
#   generates a wrapper using Python ctypes module, then defines a 
#   Python module that compiles the wrappers and Fortran source,
#   providing a Python interface to the Fortran code.
# 
#  INPUTS:
#    input_fortran_file -- str, relative or absolute path to fortran source.
# 
#  OPTIONAL:
#    name          -- str, name of output python module. Defaults to be
#                     `before_dot(os.path.basename(input_fortran_file))`
#    build_dir     -- str, the directory to store the wrapper code for
#                     `input_fortran_file`. If provided, it is
#                     assumed that the directory should not be
#                     removed automatically, otherwise a temporary
#                     directory is created, used, and deleted.
#    output_dir    -- str, the directory to store the output python
#                     module. Defaults to `os.getcwd()`.
#    dependencies  -- list of str, paths to Fortran files that should be
#                     checked for changes when compiling the final module.
#    symbols       -- list of str, names of symbols that are expected to be
#                     defined by local libraries. list of tuple of str, pairs 
#                     of symbol names and an expected substring in the library
#                     file that contains that symbol's definition.
#    **kwargs      -- Any configuration options relevant to this
#                     compilation passed as keyword arguments, 
#                     see `help(fmodpy)` for more information, or
#                     run again with 'verbose=True' to see options.
# 
#   KEYWORD OPTIONS (some important ones, more in 'fmodpy.configure'):
#    autocompile -- bool, whether or not automatic compilation of
#                   dependancies should be attempted.
#    rebuild     -- bool, True if you want to rebuild even though the
#                   `input_fortran_file` has not been modified
#                   since the last compilation of the module.
#    wrap        -- bool, True if the wrapper code should be generated
#                   even if wrapper code already exists in `build_dir`,
#                   use False when manually modifying generated wrappers.
#    show_warnings -- bool, if miscellaneous warnings should be printed.
# 
def fimport(input_fortran_file, name=None, build_dir=None,
            output_dir=None, dependencies=None, symbols=None, **kwargs):
    # Import parsing functions.
    from fmodpy.parsing import after_dot, before_dot, legal_module_name
    from fmodpy.config import run, load_config, \
        FORT_WRAPPER_EXT, PYTHON_WRAPPER_EXT, PY_EXT
    # Import "os" for os operations and "importlib" for importing a module.
    import os, sys, shutil, importlib

    # Configure this runtime of fmodpy 
    pre_config = load_config() # <- gets the configuration dictionary
    if (len(kwargs) > 0): load_config(**kwargs) # <- assigns given configuration
    # Import some locally used settings.
    from fmodpy.config import wrap, rebuild, autocompile, \
        verbose_module, f_compiler, f_compiler_args, overwrite

    # Print the configuration (when verbose).
    print()
    print("_"*70)
    print("fimport")
    print()
    print("fmodpy configuration:")
    c = load_config()
    for n in sorted(c): print(f"  {n} = {str([c[n]])[1:-1]}")
    if (len(c) > 0): del n
    del c

    # Get the source file and the source directory.
    source_path = os.path.abspath(input_fortran_file)
    source_file = os.path.basename(source_path)
    source_dir = os.path.dirname(source_path)
    

    # If the name is not defined, then try to automatically produce it.
    if (name is None): name = before_dot(source_file)

    # Check to make sure that the module name is legal.
    if not legal_module_name(name):
        from fmodpy.exceptions import NameError
        raise(NameError((f"'{name}' is not an allowed module name,\n"+
                         " must match the regexp `^[a-zA-Z_]+"+
                         "[a-zA-Z0-9_]*`. Set the name with:\n"+
                         " fmodpy.fimport(<file>, name='<legal name>')"+
                         " OR\n $ fmodpy <file> name=<legal_name>")))

    # Set the default output directory
    if (output_dir is None): output_dir = os.getcwd()
    else:                    output_dir = os.path.abspath(output_dir)

    # Initialize the list of depended files if they were not given.
    if (dependencies is None): dependencies = [source_file]
    # Given a string, assume it should be split by spaces.
    elif (type(dependencies) is str):  dependencies = dependencies.split()
    # If something else was given, then copy it into a list (because files will be appended).
    else: dependencies = list(dependencies)
    # Make sure that the source file is in the list of dependencies.
    if (source_file not in dependencies):
        dependencies.append(source_file)

    # Determine whether or not the module needs to be rebuilt.
    should_rebuild = rebuild or should_rebuild_module(
        [os.path.join(source_dir, d) for d in dependencies], name, output_dir)
    if not should_rebuild:
        print()
        print("No new modifications to '%s' module, exiting."%(name))
        print("^"*70)
        return importlib.import_module(name)

    # Generate the names of files that will be created by this
    # program, so that we can copy them to an "old" directory if necessary.
    fortran_wrapper_file = name+FORT_WRAPPER_EXT
    python_wrapper_file = name+PYTHON_WRAPPER_EXT+PY_EXT
    build_dir_name = build_dir if (build_dir is not None) else "temporary directory"
    bar_width = 23 + max(map(len,(source_dir, source_file, name, build_dir_name,
                                  fortran_wrapper_file, python_wrapper_file, output_dir)))
    print()
    print("="*bar_width)
    print("Input file directory: ",source_dir)
    print("Input file name:      ",source_file)
    print("Base module name:     ",name)
    print("Using build dir:      ",build_dir_name)
    print("  fortran wrapper:    ",fortran_wrapper_file)
    print("  python wrapper:     ",python_wrapper_file)
    print("Output module path:   ",output_dir)
    print("="*bar_width)
    print()
 
    # Prepare the build directory, link in the necessary files.
    build_dir, temp_dir = prepare_build_directory(source_dir, build_dir)

    # Find any required sources and link them into the build directory.
    if (symbols is not None):
        symbol_files = []
        print(f"Searching libraries for symbols:\n  {symbols}")
        for symbol in symbols:
            if (type(symbol) is not str):
                symbol, filename_includes = symbol
                symbol_kwargs = dict(filename_includes=filename_includes)
            else: symbol_kwargs = {}
            lib_file = load_symbol(symbol, **symbol_kwargs)
            lib_name = os.path.basename(lib_file)
            lib_dest = os.path.join(build_dir, lib_name)
            if (not os.path.exists(lib_dest)):
                print(f" sym-linking '{lib_file}' to '{lib_dest}'")
                os.symlink(os.path.realpath(lib_file), lib_dest)
            else:
                print(f" skpping sym-link to '{lib_file}', already exists at '{lib_dest}'")
            symbol_files.append(lib_name)
        print()
        dependencies = symbol_files + dependencies
    else:
        symbol_files = []

    # Automatically compile fortran files.
    if autocompile:
        print("Attempting to autocompile..")
        built, failed = autocompile_files(build_dir, dependencies)
        dependencies = dependencies + [f for f in built if (f not in dependencies)]
        print()
        if (len(built) > 0):
            print("Identified dependencies:")
            for f in built: print(" ",os.path.basename(f))
            print()

    # Write the wrappers to the files so they can be compiled.
    existing_fortran_wrapper_path = os.path.join(output_dir, name, fortran_wrapper_file)
    fortran_wrapper_path = os.path.join(build_dir, fortran_wrapper_file)
    existing_python_wrapper_path = os.path.join(output_dir, name, python_wrapper_file)
    python_wrapper_path = os.path.join(build_dir, python_wrapper_file)
    # Move the existing wrapper to the new build directory (if it is to be kept).
    if (os.path.exists(existing_fortran_wrapper_path) and (not wrap)):
        shutil.move(existing_fortran_wrapper_path, fortran_wrapper_path)
    if (os.path.exists(existing_python_wrapper_path) and (not wrap)):
        shutil.move(existing_python_wrapper_path, python_wrapper_path)
    # Check for the existence of the wrapper files (in the build or output directories).
    fortran_wrapper_exists = os.path.exists(fortran_wrapper_path)
    python_wrapper_exists = os.path.exists(python_wrapper_path)
    # Generate the wrappers for going from python <-> fortran.
    if (wrap or (not fortran_wrapper_exists) or (not python_wrapper_exists)):
        fortran_wrapper, python_wrapper = make_wrapper(
            source_path, build_dir, name)
        # Add fortran wrapper to dependencies and remove any
        #  duplicates from the "dependencies" list.
        dependencies.append( os.path.basename(fortran_wrapper_path) )
        i = len(dependencies)
        while (i > 1):
            i -= 1
            while (dependencies.index(dependencies[i]) < i):
                dependencies.pop(dependencies.index(dependencies[i]))
                i -= 1
        # Fill all the missing components of the python_wrapper string.
        python_wrapper = python_wrapper.format(
            verbose_module = verbose_module,
            f_compiler = f_compiler,
            shared_object_name = name,
            f_compiler_args = str(f_compiler_args),
            dependencies = dependencies,
            symbol_files = symbol_files,
        )
    # Write the wrapper files if this program is supposed to.
    if (not fortran_wrapper_exists) or wrap:
        with open(fortran_wrapper_path, "w") as f: f.write( fortran_wrapper )
    if (not python_wrapper_exists) or wrap:
        with open(python_wrapper_path, "w") as f: f.write( python_wrapper )

    # Make the `__init__.py` for the newly created Python module a link.
    init_file = os.path.join(build_dir,"__init__.py")
    if os.path.exists(init_file): os.remove(init_file)
    print()
    print(f"Making symlink from '{python_wrapper_file}' to '__init__.py'")
    print()
    os.symlink(os.path.join(".", python_wrapper_file), init_file)
    # Generate the final module path, move into location.
    final_module_path = os.path.join(output_dir, name)
    # Move the compiled module to the output directory.
    print("Moving from:", f"  {build_dir}", "to", f"  {final_module_path}", sep="\n")
    if not (build_dir == final_module_path):
        # Remove the existing wrapper module if it exists.
        if os.path.exists(final_module_path):
            if (overwrite):
                print(f" deleting existing directory at '{final_module_path}'..")
                shutil.rmtree(final_module_path)
            else:
                # Keep previous compilation in "old" directory.
                old_module_path = final_module_path + "_OLD"
                # Remove old directories permanently.
                if os.path.exists(old_module_path):
                    print(f" removing old wrapper of '{name}' at '{old_module_path}'..")
                    shutil.rmtree(old_module_path)
                print(f" moving existing directory at '{final_module_path}' to '{old_module_path}'..")
                shutil.move(final_module_path, old_module_path)
        # Move the compiled wrapper to the destination.
        shutil.move(build_dir, final_module_path)
        # Remove all files (symlinks) that are not dependencies.
        all_kept_files = set(dependencies) | {fortran_wrapper_file, python_wrapper_file, "__init__.py"}
        for p in os.listdir(final_module_path):
            if ((p not in all_kept_files) and (after_dot(p) != 'mod')):
                print(f" removing unnecessary file '{p}'..")
                p = os.path.join(final_module_path, p)
                os.remove(p)

    print(f"\nFinished making module '{name}'.\n")
    print("^"*70)
    print()

    # Clean up the the temporary directory if one was created.
    if temp_dir is not None:
        try: temp_dir.cleanup()
        except FileNotFoundError: pass
        del temp_dir

    # Re-configure 'fmodpy' to work the way it did before this execution.
    if (len(kwargs) > 0): load_config(**pre_config)

    # (Re)Import the module.
    sys.path.insert(0, output_dir)
    module = importlib.import_module(name)
    module = importlib.reload(module)
    sys.path.pop(0)
    # Return the module to be stored as a variable.
    return module

# ====================================================================

# Given a file name to a source python file and a build directory,
# make a python wrapper that wraps the given fortran file as if it were
# a python module. Allow for the import of all modules (with functions),
# and plain funcitons / subroutines listed in the fortran file.
def make_wrapper(source_file, build_dir, module_name):
    import os
    from fmodpy.parsing import simplify_fortran_file, after_dot
    from fmodpy.parsing.file import Fortran
    from fmodpy.exceptions import ParseError

    # Make a simplified version of the fortran file that only contains
    # the relevant syntax to defining a wrapper in python.
    is_fixed_format = after_dot(source_file) == "f"
    simplified_file = simplify_fortran_file(source_file, is_fixed_format)
    # Write the simplified file to the build directory.
    simplified_file_path = os.path.join(build_dir, "fmodpy_simplified_"+os.path.basename(source_file))
    with open(simplified_file_path, "w") as f: f.write("\n".join(simplified_file))
    # Parse the simplified fortran file into an abstract syntax.
    try:
        abstraction = Fortran(simplified_file)
    except ParseError as exc:
        from fmodpy.config import end_is_named
        if (is_fixed_format and (not end_is_named)):
            print("\nWARNING: Encountered unnamed 'END' in source for a fixed format file. Consider setting 'end_is_named=False' configuration when wrapping this file.\n")
        raise(exc)
    print("-"*70)
    print(abstraction)
    print("-"*70)
    # Generate the C <-> Fortran code.
    fortran_file = abstraction.generate_fortran()
    # Evaluate the sizes of all the Fortran variables.
    abstraction.eval_sizes(build_dir)
    # Generate the python <-> C code.
    python_file = abstraction.generate_python()
    # Return the two files that can be used to construct the wrapper.
    return fortran_file, python_file

# ====================================================================

# Given a build directory (containing some fortran files), repeatedly
# attempt to compile all files in the build directory until there are
# no successfully compiled files.
def autocompile_files(build_dir, dependencies):
    import os, time
    from fmodpy.parsing import after_dot
    # Get configuration parameters.
    from fmodpy.config import run, f_compiler, f_compiler_args, \
        wait_warning_sec, GET_SIZE_PROG_FILE, GET_SIZE_EXEC_FILE
    # Make compilation option substititions for speed.
    f_compiler_args = [arg if (arg not in {"-O3", "-O2", "-O1"}) else "-O0"
                       for arg in f_compiler_args]
    # Set the wait time and start time.
    wait_warning_sec = int(wait_warning_sec)
    start_time = time.time()
    # Try and compile the rest of the files (that might be fortran) in
    # the working directory in case any are needed for linking.
    should_compile = dependencies[:]
    dependencies = set(dependencies)
    print()
    # generate the list of files that we sould try to autocompile
    for f in os.listdir(build_dir):
        f = os.path.join(build_dir, f.strip())
        # Skip the preprocessed file, the size program, and directories
        if ( (os.path.isdir(f)) or
             ("." not in f) or
             ("f" not in after_dot(f)) or
             (GET_SIZE_PROG_FILE in f) ):
            print(f" skipping '{f}'")
            continue
        # Try opening the file, if it can't be decoded, then skip
        # Make sure the file does not have any immediate exclusions,
        # if it does then skip it
        try:
            with open(f) as fort_file:
                f = os.path.basename(f)
                print(f" reading '{f}' to check if it can be autocompiled.. ", end="")
                exclude_this_file = False
                # Read through the file, look for exclusions
                for line in fort_file.readlines():
                    line = line.strip().upper().split()
                    if (len(line) > 0) and ("PROGRAM" in line[0]):
                        exclude_this_file = True
                        break
                if exclude_this_file:
                    print(f"no. The file '{f}' contains 'PROGRAM'.")
                    continue
                else: print("yes.")
        # Some failure occurred while reading that file, skip it
        except UnicodeDecodeError: continue
        # No failures or obvious red-flags, this file might be useful
        should_compile.append(f)
    # Handle dependencies by doing rounds of compilation, presuming
    # only files with fewest dependencies will compile first
    ordered_depends = []
    successes = {None}
    made_mod = set()
    failed = set()
    # Get the list of existing "mod" files (Fortran module definitions).
    current_mod_files = sum((1 for f in os.listdir(build_dir) if after_dot(f) == "mod"))
    previous_mod_files = current_mod_files
    # Continue rounds until (everything compiled) or (no successes nor new mod files)
    print()
    while (len(should_compile) > 0) and (
            (len(successes) > 0) or (len(made_mod) > 0)):
        successes = set()
        made_mod = set()
        for f in should_compile:
            # Check to see if compilation is taking unexpectedly long.
            if (time.time()-start_time > wait_warning_sec):
                import warnings
                warnings.warn("\n  Automatic compilation is taking longer than expected, consider"+
                              "\n  providing explicitly ordered dependencies (precompiled or not) with"+
                              "\n    fmodpy.fimport('<target>', dependencies=['<path>', ...], ...)\n")
                start_time = float('inf') # <- do this to prevent further warnings

            # Try to compile all files that have "f" in the extension.
            print(f"Compiling '{f}'.. ")
            # Reuse the "GET_SIZE_EXEC_FILE" name for compilation checks.
            compiled_file_path = os.path.join(build_dir, GET_SIZE_EXEC_FILE)
            cmd = [f_compiler] + f_compiler_args + ["-o", compiled_file_path] + \
                  [d for d in ordered_depends if (d != f)] + [f]
            print(f" {' '.join(cmd)}".replace(build_dir,"."))
            code, stdout, stderr = run(cmd, cwd=build_dir)
            # Update the list of existing mod files.
            previous_mod_files = current_mod_files
            current_mod_files = sum((1 for f in os.listdir(build_dir) if after_dot(f) == "mod"))
            if (code == 0):
                successes.add(f)
                if (f not in ordered_depends):
                    ordered_depends.append(f)
                print("  success.")
                # If all dependencies have been successfully
                #  compiled, then exit (because we are done).
                if (f in dependencies):
                    dependencies.remove(f)
                if (len(dependencies) == 0):
                    should_compile = []
                # Since a file was successfully compiled, break, which
                #  will trigger another attempt at compiling the
                #  target file (if there is a target).
                break
            elif (current_mod_files > previous_mod_files):
                print("  failed, but created new '.mod' file, continuing.")
                made_mod.add(f)
                if (f not in ordered_depends):
                    ordered_depends.append(f)
            else:
                # Record failed compilations.
                if (f not in failed): failed.add(f)
                print("  failed.")
                print("NAME:", f)
                
                if (max(len(stdout), len(stderr)) > 0): print('-'*70)
                if len(stdout) > 0:
                    print("STANDARD OUTPUT:")
                    print("\n".join(stdout))
                if len(stderr) > 0:
                    stderr_str = "\n".join(stderr)
                    print("STANDARD ERROR:")
                    print(stderr_str)
                    # If there is a pure "error" in the target file, raise it.
                    if ("\nError:" in stderr_str) and (f in dependencies):
                        from fmodpy.exceptions import CompileError
                        raise CompileError(f"Failed to compile '{f}' with error:\n\n{stderr_str}")
                if (max(len(stdout), len(stderr)) > 0): print('-'*70)
        # Remove the files that were successfully compiled from
        # the list of "should_compile" and the list of "failed".
        for f in successes:
            if (f in failed):
                failed.remove(f)
            # The only way that "f" could not be in "should_compile"
            #  is if all dependencies were compiled successfully and
            #  "should_compile" was overwritten with an empty list.
            if (f not in made_mod) and (len(should_compile) > 0):
                should_compile.remove(f)
    # Log the files that failed to compile.
    for f in should_compile: print(f"Failed to compile '{f}'.")
    # Raise an error if there was a target file and it was not compiled.
    failed_dependencies = dependencies & failed
    if (len(failed_dependencies) > 0):
        from fmodpy.exceptions import CompileError
        raise(CompileError(f"Failed to compile {failed_dependencies}.\n  Current compilation arguments are:\n    {f_compiler_args}\n  Is a necessary compilation argument or dependency missing?"))
    # Return the list of files that were successfully compiled in
    # the order they were compiled and the files that failed to compile.
    return ordered_depends, failed

# ====================================================================

# Given a symbol name, try to locate the library that defines it
#   and load the symbol as a CDLL so that it is accessible.
def load_symbol(symbol, filename_includes=""):
    import os, subprocess, re, ctypes
    from fmodpy.config import libraries, library_recursion, library_extensions, symbol_command
    # This code will find the first match for the symbol in a library.
    max_recursion = library_recursion
    versions = set()
    directories = [(p,max_recursion) for p in libraries]
    checked_files = 0
    symbol_file = None
    while ((len(directories) > 0) and (symbol_file is None)):
        directory, max_recursion = directories.pop(0)
        # Skip nonexistant directories.
        if (not os.path.exists(directory)):
            print(f" MISSING - no such path '{directory}'.")
            continue
        if (not os.access(directory, os.R_OK)):
            print(f" SKIPPING - not allowed to read '{directory}'.")
            continue
        # Get all files with the desired extension.
        file_paths = [
            os.path.join(directory, f)
            for f in os.listdir(directory)
            if  (not os.path.isdir(os.path.join(directory, f)))
            and (any(ext in library_extensions for ext in f.split(".")[1:]))
            and (filename_includes in f)
        ]
        # Iterate over candidate files and check their symbol tables.
        for f in file_paths:
            checked_files += 1
            print(f" checking for '{symbol}' in '{f}' with '{symbol_command.format(path=f)}'.")
            symbols = str(subprocess.run(symbol_command.format(path=f), shell=True,
                                         capture_output=True).stdout, "utf8")
            if (symbol+"\n" in symbols):
                symbol_file = f
                break
        # Recurse into directories (up to provided max depth).
        if ((max_recursion > 0) and (symbol_file is None)):
            directories += [
                (os.path.join(directory, d), max_recursion-1)
                for d in os.listdir(directory)
                if os.path.isdir(os.path.join(directory, d))
            ]
    # Raise an error if no symbol file was found.
    if (symbol_file is None):
        raise(NotImplementedError(f"After checking {checked_files} files, the symbol '{symbol}' could not be found in any of the fmodpy.config.libraries paths:\n   {libraries}"))
    # Load the library globally (so it is accessible to later codes) and return its path.
    ctypes.CDLL(symbol_file, mode=ctypes.RTLD_GLOBAL)
    return symbol_file

# ====================================================================

# Given the path to the file that we are creating an extension for,
# create and prepare a build directory for the project compilation
def prepare_build_directory(source_dir, build_dir):
    import os
    # Create a build directory.
    if (build_dir is None):
        # Create a temporary directory for building.
        from tempfile import TemporaryDirectory
        temp_dir = TemporaryDirectory()
        build_dir = temp_dir.name
        print(f"Using temporary build directory at '{build_dir}'.")
    else:
        # Otherwise, assume a path was given, convert to absolute form.
        temp_dir = None
        build_dir = os.path.abspath(build_dir)
        # Create the directory for the build if it does not exist.
        if (not os.path.exists(build_dir)):
            print(f"Making build directory at '{build_dir}'.")
            os.makedirs(build_dir) #, exist_ok=True)

    # If the working directory is not the same as the file directory,
    # copy all the contents of the file directory into the working
    # directory (in case any of them are used by the fortran project)
    if (os.path.abspath(source_dir) != os.path.abspath(build_dir)):
        print("Build directory is different from source directory..")
        for f in os.listdir(source_dir):
            # Get the full path of this source file.
            source = os.path.join(source_dir,f)
            # Do not make links to python files (because they shouldn't be needed)
            if (f[-3:] == ".py"): continue
            # Do not make links to the directory itself.
            if (source == build_dir): continue
            # Create a symbolic link to all source files in the build directory.
            destination = os.path.join(build_dir,f)
            print(" sym-linking '%s' to '%s'"%(source, destination))
            # Remove existing symbolic links (in case they are dead).
            if (os.path.islink(destination)): os.remove(destination)
            # Skip existing files (copied in manually, do not delete).
            elif (os.path.exists(destination)): continue
            # Create a new symbolic link.
            os.symlink(source, destination)

    print()
    # Return the prepared working directory
    return build_dir, temp_dir

# ====================================================================

# Return True if a module should be rebuilt, False otherwise.
def should_rebuild_module(dependencies, module_name, module_directory):
    # Get the modification time of the Python package file.
    import os, sys, pkgutil, importlib
    # Get the last modification time of the module (if it exists already)
    # Make sure that the output directory is in the sys path so that
    # the time-check import loader will work correctly.
    sys.path.insert(0, module_directory)
    package = pkgutil.get_loader(module_name)
    if package is None: module_mod_time = 0
    else:               module_mod_time = os.path.getmtime(package.path)
    # Get the latest modification time of all source files.
    source_mod_time = max(map(os.path.getmtime, dependencies))
    # Return `True` if the source file has been modified since the
    # last construction of the module.
    if (module_mod_time <= source_mod_time):
        sys.path.pop(0)
        return True
    # If the module file is newer than the source file, try to import.
    try: importlib.import_module(module_name)
    except ImportError: return True # problem with the built module..
    finally: sys.path.pop(0)
    # The module successfully imported, no rebuild necessary.
    return False

# ====================================================================

# Save a configuration globally on this machine. To see the list of
# all configurations, call with no arguments.
def configure(*to_delete, **to_save):
    # Get the "home_directory" and the "config_file" for saving.
    import os, time, builtins
    from fmodpy.config import home_directory, config_file
    # Use the built-in print function, instead of the fmodpy one (default).
    from builtins import print
    # Otherwise, add the given configuration values to the config file.
    path = os.path.join(home_directory, config_file)
    # Read the existing file, if it's there.
    lines = []
    if (os.path.exists(path)):
        with open(path, "r") as f:
            lines = [l.strip() for l in f.readlines()]
    # If no arguments were given, print the current
    # configuration to standard output.
    if ((len(to_delete) == 0) and (len(to_save) == 0)):
        import fmodpy
        from fmodpy.config import load_config
        existing = {''.join([v.strip() for v in l.split("=")[:1]]) for l in lines}
        conf = load_config()
        # Collect the lines of the printout.
        lines = [f"fmodpy ({fmodpy.__version__}):"]
        finished_existing = False
        if (len(existing) > 0): lines += [f" # configuration declared in {path}"]
        for name in sorted(sorted(conf), key=lambda n: n not in existing):
            if (len(existing) > 0) and (name not in existing) and (not finished_existing):
                lines += [""," # default values"]
                finished_existing = True
            lines.append(f"  {name} = {str([conf[name]])[1:-1]}")
        # Print the output.
        print("\n".join(lines+[""]))
    else:
        # Overwrite the file, commenting out all the old stuff.
        with open(path, "w") as f:
            to_remove = {n for n in to_delete}.union(set(to_save))
            # Write the old contents, commented out.
            for l in lines:
                # If this variable is being overwritten, comment out the former.
                comment_out = any(v.strip() in to_remove for v in l.split("=")[:1])
                if comment_out: l = "# "+l
                # Add the line to the file (so there is a history).
                print(l, file=f)
            # Write the new contents.
            print("", file=f)
            print("# ",time.ctime(), file=f)
            for name in sorted(to_save):
                value = to_save[name]
                str_value = str([to_save[name]])[1:-1]
                print(f"{name} = {str_value}", file=f)



from . import class_name

# Base class for all blocks of Fortran code that will be wrapped. This
# information describes the code in a common format.
class Code:
    # Description
    name = ""
    type = None
    docs = ""
    lines = 0
    allowed_unknown = False
    # Does the end of this Code have to be named?
    needs_end = True
    named_end = True
    # Parent module.
    parent = None
    # Declare some naming convention for parsing
    prefixes = []
    # List of (func, str) pairs where the string is the local attribute
    # name and the function is a parser returning a list of objects.
    can_contain = []
    # List of (func) parsing functions that handle ignorable contents.
    will_ignore = []

    # Initialize this object.
    def __init__(self, list_of_lines, preceding_comments="", parent=None):
        self.parent = parent
        self.docs = preceding_comments
        # Initizlie lists of contained objects.
        for (_, name) in self.can_contain: setattr(self, name, [])
        # Parse the rest of this abstract object from file.
        # Count how many lines this object has by tracking change.
        starting_length = len(list_of_lines)
        self.parse(list_of_lines)
        self.lines += (starting_length - len(list_of_lines))

    # Given a list of strings (lines of simplified Fortran file),
    # where then entire string has been converted to upper case,
    # initialize this object.
    def parse(self, list_of_lines):
        from fmodpy.config import fmodpy_print as print
        print(f"{self.type.title()}.parse {self.name}")
        # Extract documentation first.
        while (len(list_of_lines) > 0):
            stripped_line = list_of_lines[0].strip() + '!'
            if (stripped_line[0] == '!'):
                self.docs += "\n" + list_of_lines.pop(0)
            # If this is not a comment line, the docs are done.
            else: break
        # Strip off any whitespace from the edges of the docs.
        self.docs = self.docs.strip()
        # Strip off all empty comment lines from the end of the comment block.
        while (self.docs[-2:] == "\n!"): self.docs = self.docs[:-2]
        # Parse the rest of the file (popping out lines from list).
        ended = False
        comments = ""
        while (len(list_of_lines) > 0):
            line = list_of_lines[0].strip().split()
            if len(line) == 0:
                comments = "" # empty new lines before something indicate no comment connection
                list_of_lines.pop(0)
                continue
            # Check if this is a comment (line should not be empty).
            elif (line[0][:1] == "!"):
                comments += "\n" + list_of_lines.pop(0)
                continue
            # Check to see if this is the END of this object.
            elif (line[0] == "END"):
                # If this line only contains the keyword "END" ...
                if (len(line) <= 1):
                    from fmodpy.config import end_is_named
                    if end_is_named:
                        from fmodpy.exceptions import ParseError
                        raise(ParseError("Encountered unexpected 'END' without block type (e.g. SUBROUTINE, MODULE)."))
                    else:
                        list_of_lines.pop(0)
                        ended = True
                        break
                # Otherwise, check what is ending ...
                elif (line[1] == self.type):
                    if (self.named_end):
                        if (len(line) < 3):
                            from fmodpy.exceptions import ParseError
                            raise(ParseError(f"Encountered unexpected '{list_of_lines[0].strip()}' without named ending."))
                        elif (line[2] == self.name):
                            list_of_lines.pop(0)
                            ended = True
                            break
                    # This does not require a named ending, we are done.
                    else:
                        list_of_lines.pop(0)
                        ended = True
                        break
                # This is the end of something else (not me), ignore it.
                else:
                    list_of_lines.pop(0)
                    continue
            # Parse the contained objects out of the file.
            for (parser, name) in self.can_contain:
                pre_length = len(list_of_lines)
                instances = parser(list_of_lines, comments, self)
                length = pre_length - len(list_of_lines)
                self.lines += length
                # If instances were found, we have a match, break.
                if (len(instances) > 0):
                    for inst in instances:
                        getattr(self,name).append( inst )
                    comments = ""
                    break
                # If any parsing was done otherwise, then it was a match, break.
                elif (length > 0): break
            else:
                # Look for things that will be parsed, but ignored.
                for parser in self.will_ignore:
                    pre_length = len(list_of_lines)
                    instances = parser(list_of_lines, comments, self)
                    length = len(list_of_lines) - pre_length
                    self.lines += length
                    if ((len(instances) > 0) or (length > 0)): break
                else:
                    # This is an unknown block of code.
                    if self.allowed_unknown:
                        comments = ""
                        # This is an un-identifiable line, it belongs ???
                        print(f"  skipping line '{list_of_lines.pop(0)}'")
                    else:
                        from . import class_name
                        from fmodpy.exceptions import ParseError
                        raise(ParseError(f"\n\nEncountered unrecognized line while parsing {class_name(self)}:\n\n  {str([list_of_lines.pop(0)])[1:-1]}\n\n"))
        # Check for correct ending.
        if (self.needs_end and (not ended)):
            from fmodpy.exceptions import ParseError
            raise(ParseError(f"File ended without observing 'END {self.type}'."))
        # Finalize docs (knowing they will be formatted soon).
        self.docs = self.docs.replace("{","{{").replace("}","}}")
        # Denote the end of the docs.
        print(f" {self.type.title()}.parse done.")


    # Parse the contained objects into Fortran BIND(C) code.
    def generate_fortran(self):
        lines = []
        for (_, name) in self.can_contain:
            for instance in getattr(self, name):
                lines += instance.generate_fortran()
        return lines

    # Parse the contained objects into python code lines, making
    # Python-callable code for all contained objects in this Code.
    def generate_python(self):
        lines = []
        type_blocks = set()
        for (_, name) in self.can_contain:
            for instance in getattr(self, name):
                # TODO: Find a way to generalize this pattern better,
                #       should not need custom logic like this.
                #       Right now it collects types from all children
                #       then adds those type definitions to self.
                args = []
                if (instance.type in {"SUBROUTINE","FUNCTION","MODULE"}):
                    args.append(type_blocks)
                # Get the lines of python code.
                lines += instance.generate_python(*args) + ['']
        # Add the type declarations if there are any.
        for block in sorted(type_blocks):
            lines.insert(0,"")
            for line in reversed(block.split("\n")):
                lines.insert(0, line)
        # Return the full list of lines.
        return lines

    # Evaluate the btye-width of all contents (by compiling Fortran test).
    def eval_sizes(self, build_dir):
        # Recursively evaluate the sizes of all children.
        for (_, name) in self.can_contain:
            for instance in getattr(self, name):
                if (hasattr(instance, "eval_sizes")):
                    instance.eval_sizes(build_dir)
        # After evaluating the sizes for all children, if this code object
        # has arguments inside of it, evaluate their sizes.
        if hasattr(self, "arguments"):
            # Evaluate the size of all arguments for this routine.
            size_prog =  "PROGRAM GET_SIZE\n"
            # Add a use line for the module this is in (if it is inside one).
            if (self.parent is not None):
                # If the parent is a module, assume access to it with a USE.
                if (self.parent.type == "MODULE"):
                    size_prog += f"  USE {self.parent.name}\n"
                # Add any used modules by the parent (because this has access to those).
                if (hasattr(self.parent, "uses")):
                    for line in self.parent.uses: size_prog += f"  {line}\n"

            # # If this is a TYPE, then add the type declaration.
            # if (self.type == "TYPE"):
            #     size_prog += "\n".join(["  " + l for l in str(self).split("\n")]) + "\n"

            # If this is not a TYPE, then it might have USES statements.
            if hasattr(self, "uses"):
                for line in sorted(self.uses): size_prog += line+"\n"
            # Get the unique argument type:kind pairs, all arguments as values.
            unique_types_and_kinds = {}
            for arg in self.arguments:
                # Skip "TYPE" arguments, their sizes are evaluated separately.
                if (arg.type == "TYPE"): continue
                # Append a unique identifier to "RESULT" for functions.
                if (hasattr(self, "result") and (arg.name == self.name)):
                    arg.name += "_result"
                key = (arg.type,arg.kind)
                unique_types_and_kinds[key] = unique_types_and_kinds.get(key,[]) + [arg]
            # Add all unique (type,kind) arguments to the size check program.
            for (t,k) in sorted(unique_types_and_kinds):
                # Create a singleton argument of the same type to measure size.
                arg = unique_types_and_kinds[(t,k)][0]
                temp_arg = type(arg)([arg.type])
                temp_arg.name = arg.function_safe_name()
                temp_arg.kind = arg.kind
                temp_arg.show_intent = False
                size_prog += f"  {temp_arg}\n"
            # Add print statements (printing the sizes).
            for k in sorted(unique_types_and_kinds):
                name = unique_types_and_kinds[k][0].name
                size_prog += f"  WRITE (*,*) SIZEOF({name})\n"
            # End the program file.
            size_prog += "END PROGRAM GET_SIZE\n"
            # Take the size program string and make sure all lines 
            #  fit within the expected character width.
            from fmodpy.parsing.util import wrap_long_lines
            size_prog = "\n".join(wrap_long_lines(size_prog.split("\n")))
            # Import some necessary configurations to make the SIZEOF program.
            from fmodpy.config import run, f_compiler, \
                GET_SIZE_PROG_FILE, GET_SIZE_EXEC_FILE
            import os
            size_prog_path = os.path.join(build_dir, GET_SIZE_PROG_FILE)
            size_exec_path = os.path.join(build_dir, GET_SIZE_EXEC_FILE)
            # Write the size program.
            with open(size_prog_path, "w") as f: f.write(size_prog)
            # Compile an executable "GET_SIZE" program. All of the 
            # necessaray modules (for typing) should already be compiled.
            code, stdout, stderr = run([f_compiler, "-o", size_exec_path, size_prog_path])
            if (code != 0):
                from fmodpy.exceptions import CompileError
                error_message = "\n\nCompilation of a program for size checking failed.\n"+\
                                     "This may have happened because necessary module files\n"+\
                                     "are not compiled and present in the build directory.\n\n"+\
                                "\n".join(stderr)
                raise(CompileError(error_message))
            # Run the "GET_SIZE" executable.
            code, stdout, stderr = run([size_exec_path])
            if (code != 0): raise(NotImplementedError("Argument byte-size checking program failed, unexpected."))
            # Remove the size program.
            os.remove(size_prog_path)
            os.remove(size_exec_path)
            # Read the sizes from the output.
            sizes = [s.strip() for s in stdout if s.strip().isnumeric()]
            # Assign the sizes to the appropriate arguments.
            for s, key in zip(sizes, sorted(unique_types_and_kinds)):
                for arg in unique_types_and_kinds[key]:
                    arg.size = s



from .code import Code
from .argument import Argument
from . import parse_argument

STRUCT_DEFAULT_VALUE = 0
CONTENTS_DECLARATION = '("{n}", {t})'
FIELD_DECLARATION = 'self.{n} = kwargs.get("{n}", getattr(value, "{n}", value))'
STRUCTURE_DEFINITION = '''
# This defines a C structure that can be used to hold this defined type.
class {name}(ctypes.Structure):
    # (name, ctype) fields for this structure.
    _fields_ = [{fields}]
    # Define an "__init__" that can take a class or keyword arguments as input.
    def __init__(self, value={default}, **kwargs):
        # From whatever object (or dictionary) was given, assign internal values.
        {field_declarations}
    # Define a "__str__" that produces a legible summary of this type.
    def __str__(self):
        s = []
        for (n, t) in self._fields_:
            s.append( n + "=" + str(getattr(self,n)) )
        return "{name}[" + ", ".join(s) + "]"
    # Define an "__eq__" method that checks equality of all fields.
    def __eq__(self, other):
        for (n, t) in self._fields_:
            if (getattr(self, n) != getattr(other, n, None)):
                return False
        return True
'''


class TypeDeclaration(Code):
    type = "TYPE"
    can_contain = [(parse_argument, "arguments")]

    # Given a list of lines (of a source Fortran file), parse out this
    # TYPE (assuming the first line is the line declaring this
    # Subroutine).
    def parse(self, list_of_lines):
        # If there is "TYPE (" then this doesn't declare a type, it is an argument.
        if ("(" == list_of_lines[0][5:6]): return
        # Otherwise, this declares a type.
        self.lines += 1
        if ('BIND ( C )' not in list_of_lines[0]):
            from fmodpy.exceptions import NotSupportedError
            raise NotSupportedError(f"\n  Fortran derived types must be C compatible and include 'BIND(C)' to be wrapped.\n  Encountered this issue when parsing the following line:\n    {list_of_lines[0]}")
        # Parse the name of this subroutine out of the argument list.
        declaration_line = list_of_lines.pop(0).strip().split()
        assert (':' in declaration_line), "Expected ':' to exist in TYPE declaration line:\n  {' '.join(declaration_line)}"
        colon_index = -declaration_line[::-1].index(':')-1
        declaration_line = declaration_line[colon_index+1:]
        # Extract any "arguments" to the type.
        argument_names = set()
        #   Assume the format "TYPE :: name"
        #                  or "TYPE :: name(arg1, arg2, ...)"
        if ('(' in declaration_line):
            self.name = declaration_line.pop(0)
            argument_names = {
                token for token in declaration_line
                if token not in {'(', ',', ')'}
            }
        else:
            self.name = declaration_line[-1]
        # ------- Default parsing operations -------
        super().parse(list_of_lines)
        # ------------------------------------------
        # Disable showing "INTENT" on all arguments of this TYPE.
        for arg in self.arguments:
            arg.show_intent = False
        # Verify that all arguments are defined and permissable.
        if (len(argument_names) > 0):
            for arg in self.arguments:
                if (arg.name in argument_names):
                    if (arg.type_kind):
                        from fmodpy.exceptions import NotSupportedError
                        raise(NotSupportedError(
                            "\nDerived types with KIND parameters are not supported yet."+
                            "\n  Consider posting a simplified example with your motivation to"+
                            "\n  https://github.com/tchlux/fmodpy/issues"
                        ))
                    elif (not arg.type_len):
                        raise(NotImplemented())
                    argument_names.remove(arg.name)
            assert (len(argument_names) == 0), f'TYPE arguments {argument_names} were not defined within the type.'


    # Declare the lines that create the corresponding Struct for this type.
    # TODO: Treat the type like it is a c function being called, include
    #       size parameters for arrays (when it is not C compatible).
    def py_declare(self):
        # Get the names and types necessary to declare a C-compatible struct.
        name = self.name
        names = [arg.name for arg in self.arguments]
        types = [arg.c_type for arg in self.arguments]
        struct = STRUCTURE_DEFINITION.format(
            name = name,
            default = STRUCT_DEFAULT_VALUE,
            fields = ", ".join([CONTENTS_DECLARATION.format(n=n.lower(), t=t)
                                for (n,t) in zip(names, types)]),
            field_declarations = "\n        ".join([FIELD_DECLARATION.format(n=n.lower())
                                                    for n in names])
        )
        # Return the list of lines that defines this struct.
        return struct.split("\n")

    # Declare the lines that create this type in Fortran.
    def fort_declare(self):
        return str(self).split("\n")

    def __str__(self):
        # Define the arguments to the type (for constructing its internals).
        type_args = [a.name for a in self.arguments if a.type_len]
        if (len(type_args) > 0):
            args = f"(" + ", ".join(type_args) + ")"
        else:
            args = ""
        # Define the type.
        out = f"{self.type}, BIND(C) :: {self.name}{args}\n" 
        # Add arguments.
        for a in self.arguments:
            temp = a.copy()
            temp.show_intent = False
            out += f"  {temp}\n"
        # End type.
        out += f"END {self.type} {self.name}"
        return out


class TypeArgument(Argument):
    type = "TYPE"
    c_types = {}
    default_singleton = 1
    kind_prefix = ""

    # For a type, the "c_type" in python is the type name (assuming
    #  the struct was defined in an accessible scope).
    @property
    def c_type(self):
        return self.kind

    # TODO: Make both the type and a C compatible equivalent.
    # def fort_declare(self):

    # TODO: Copy the input into the true Fortran type.
    # def fort_before(self): 

    # TODO: Copy the result from the call outputs into the C inputs.
    # def fort_after(self): 
